#!/usr/bin/python
# -*- coding: utf8 -*-

"""
Disney Disparity.
:author: Manuel Tuschen
:date: 20.06.2016
:license: GPL3
"""

from __future__ import division, absolute_import, unicode_literals, print_function

import os
import argparse

import numpy as np
import h5py
from skimage.io import imread
from skimage.util import img_as_float, img_as_ubyte, img_as_uint
from skimage.color import rgb2gray, gray2rgb
from easyScripting import multiLoading, prepareSaving




def imgs2lf(input_dir, output_file, output_dataset='lightfield', img_extension = '.png', dtype=np.uint8, RGB=True):
    """
    Convert several images to a lightfield.

    Parameters
    ----------
    input_dir : string
        The directory where the ligthfield images are located.
    output_file: string
        The filename  (including the directory), of the output file.
    output_dataset : string, optional
        The new container name of the hdf5 file.
    img_extension : string, optional
        The file extension of the images to look for.
    dtype : numpy.dtype, optional
        The new data type for the downscaled lightfield. Must be either
        np.float64, np.uint8 or np.uint16.
    RGB : bool, optional
        If True, the output lightfield will be converted to RGB (default).
        Otherwise gray type images are stored.
    """

    # look for images
    files = multiLoading(directory=input_dir, extension=img_extension)

    # prepare saving
    output_name = os.path.basename(output_file)
    output_dir = os.path.dirname(output_file)
    output_file = prepareSaving(output_name, path=output_dir, extension=".hdf5")


    # Which dtye should be used?
    if dtype is np.float64:
        img_0 = img_as_float(imread(files[0]))
    elif dtype is np.uint8:
        img_0 = img_as_ubyte(imread(files[0]))
    elif dtype is np.uint16:
        img_0 = img_as_uint(imread(files[0]))
    else:
        raise TypeError('Given dtype not supported.')


    rows, cols, orig_channels = img_0.shape # automatically determine the images'shapes from the first image.


    # Do we shall take RGB or gray images?
    if orig_channels != 3 or orig_channels != 1:
        raise TypeError('The given images are neither gray nor RGB images')
    if RGB:
        channels = 3
    else:
        channels = 1


    f_out = h5py.File(output_file, 'w')
    dataset = f_out.create_dataset(output_dataset, shape=(len(files), rows, cols, channels), dtype=dtype)
    for k in range(len(files)):

        if dtype is np.float64:
            if orig_channels==1 and RGB:
                dataset[k, ...] = img_as_float(gray2rgb(imread(files[k])))
            elif orig_channels==3 and not RGB:
                dataset[k, ...] = img_as_float(rgb2gray(imread(files[k])))
            else:
                dataset[k, ...] = img_as_float(imread(files[k]))
        elif dtype is np.uint8:
            if orig_channels==1 and RGB:
                dataset[k, ...] = img_as_ubyte(gray2rgb(imread(files[k])))
            elif orig_channels==3 and not RGB:
                dataset[k, ...] = img_as_ubyte(rgb2gray(imread(files[k])))
            else:
                dataset[k, ...] = img_as_ubyte(imread(files[k]))
        elif dtype is np.uint16:
            if orig_channels==1 and RGB:
                dataset[k, ...] = img_as_uint(gray2rgb(imread(files[k])))
            elif orig_channels==3 and not RGB:
                dataset[k, ...] = img_as_uint(rgb2gray(imread(files[k])))
            else:
                dataset[k, ...] = img_as_uint(imread(files[k]))
        else:
            raise TypeError('Given dtye not supported.')
    f_out.close()




################################################################################
#                                                                              #
#                       Can be used as a command line tool                     #
#                                                                              #
################################################################################



if __name__ == "__main__":

    parser = argparse.ArgumentParser(description='Store image files into hdf5 container.')

    parser.add_argument('input_dir', help='The directory where the lightfield images are located.')
    parser.add_argument('--output_file', help='The filename (including the directory) of the output .hdf5 file.', default='lightfield')
    parser.add_argument('--output_dataset', help='The container name in the hdf5 file.', default='lightfield')
    parser.add_argument('--img_extension', help='The file extension of the images to look for.', default='.png')
    parser.add_argument('--dtype', help='The dtype used to store the lightfild data.', choices=['float', 'uint', 'ubyte'], default='uint')
    parser.add_argument('-RGB', help='Flag to determine if resulting lightfield should consists of RGB values.', action='store_true')

    args = parser.parse_args()

    type = np.float64
    if args.dtype == 'ubyte':
        type = np.uint8
    if args.dtype == 'uint':
        type = np.uint16

    imgs2lf(args.input_dir, output_file=args.output_file, output_dataset=args.output_dataset, img_extension=args.img_extension, dtype=type, RGB=args.RGB)


#!/usr/bin/python
# -*- coding: utf8 -*-

"""
Disney Disparity.
:author: Manuel Tuschen
:date: 20.06.2016
:license: GPL3
"""

from __future__ import division, absolute_import, unicode_literals, print_function

import warnings

import numpy as np
from skimage.morphology import binary_opening, square




def edge_confidence(epi, window=9, threshold=0.02):
    """
    Calculates the edge confidence according to eq. 2.
    This is a simple measurement which designates an edge if pixel intensities
    of gray value changes. An edge is determined by the sum of difference in
    pixel intensity of a central pixel to all pixels in a 1D window of size
    window. It is assumed that there is an edge if the sum is greater than a
    given threshold.

    Parameters
    ----------
    epi : numpy.array [v,u]
        Set of all gray-value epis for scanline s_hat.
    window_size : int, optional
        The 1D window siz in pixels. As the window should be centered
        symmetrically around the pixel to process the value shpuld be odd. For
        even numbers the next higher odd number is chosen.
    threshold : float, optional
        The threshold giving the smallest difference in EPI luminescence which
        must be overcome to designate an edge.

    Returns
    -------
    Ce : numpy.array [v,u]
         Edge confidence values for each EPI pixel.
    Me : numpy.array [v,u] of boolean.
        True means an edge was discovered for at that pixel.
    """

    v_dim = epi.shape[0]
    u_dim = epi.shape[1]
    # Check dimensions of input data
    assert epi.shape == (
        v_dim,
        u_dim,), 'Input EPI has wrong shape in function \'edge_confidence\'.'

    # Make window size odd
    if window % 2 == 0:
        warnings.warn(
            'window should be an odd number in function \'edge_confidence\'. Window size {g} was given but {u} is used instead.'.format(
                g=window, u=window + 1))
        window += 1

    # We avoid the border problem by padding the epi.
    padded_epi = np.pad(epi, ((0, 0), (int(window // 2), int(window // 2))),
                        'edge')
    assert padded_epi.shape == (v_dim,
                                u_dim + window - 1,), 'Padded epi has wrong shape in function \'edge_confidence\'.'

    # Calculate confidence values
    Ce = np.zeros(epi.shape, dtype=np.float64)  # initiate array
    for k in range(window):
        Ce += (epi[...] - padded_epi[:, k:epi.shape[1] + k]) ** 2
    Me = Ce > threshold  # create confidence Mask
    Me = binary_opening(Me, selem=square(2),
                        out=Me)  # work with square to avoid aliasing

    # Let's see if our results have reasonable meaning'
    assert np.all(
        Ce >= 0), 'Negative edge confidence found in function \'edge_confidence\'.'
    assert Ce.shape == (v_dim,
                        u_dim,), 'Ce output has incorrect shape in fucntion \'edge_confidence\'.'
    assert Me.shape == (v_dim,
                        u_dim,), 'Me output has incorrect shape in fucntion \'edge_confidence\'.'

    return Ce, Me


def disparity_confidence(S, Ce, threshold=0.1):
    """
    Calculate the disparity confidence according to eq. 7 Disparity confidence
    is calculated for each pixel u in one line s from an EPI by combining the
    estimated scores found during radiance sampling and the edge confidence
    for each pixel.

    Parameters
    ----------
    S : numpy.array [u,d]
        The calculated scores for each sampled radiance at each pixel u.
    Ce :  numpy.array [u]
         Edge confidence values for each EPI pixel.
    threshold : float, optional
        The threshold which must be overcome for the depth confidence.

    Returns
    -------
    Cd :  numpy.array [u]
         The calculated confidence values for each pixel.
    Md : numpy.array [u] of boolean.
        True means edge confidence is sufficient.
    S_max : numpy.array [u]
        The maximal score over all disparities per pixel.
    S_mean : numpy.array [u]
        The mean score over all disparities per pixel.
    """

    u_dim = S.shape[0]
    d_dim = S.shape[1]
    # Check dimensions of input data
    assert S.shape == (u_dim,d_dim), 'Input S has wrong shape in function \'disparity_confidence\'.'
    assert Ce.shape == (u_dim,), 'Input Ce has wrong shape in function \'disparity_confidence\'.'

    S_max = np.max(S, axis=-1)
    S_mean = np.sum(S, axis=-1)  # might be np.mean but is written as sum in the paper
    Cd = Ce * abs(S_max - S_mean)

    Md = Cd > threshold  # create confidence Mask

    # Let's see if our results have reasonable meaning'
    assert np.all(
        Cd >= 0), 'Negative disparity confidence found in function \'disparity_confidence\'.'
    assert Cd.shape == (
        u_dim,), 'Cd output has incorrect shape in fucntion \'disparity_confidence\'.'
    assert Md.shape == (
        u_dim,), 'Md output has incorrect shape in fucntion \'disparity_confidence\'.'

    return Cd, Md, S_max, S_mean



#!/usr/bin/python
# -*- coding: utf8 -*-

"""
Disney Disparity.
:author: Manuel Tuschen
:date: 20.06.2016
:license: GPL3
"""

from __future__ import division, absolute_import, unicode_literals, print_function

import os
import numpy as np

from skimage.io import imsave

from ._confidence import disparity_confidence
from ._radiance import sample_radiance, find_disp_bounds
from ._score import score_computation
from ._disparity import estimate_disp


def process_epi(epi, D, Ce, M, s_hat, min_disp, max_disp, stepsize, Cd_t,
                NOISEFREE, COARSEST, DEBUG=False, s_hat_DEBUG=0,
                DEBUG_dir='/tmp', v_DEBUG=0):
    """
    Depth estimates according to eq. 6 for one scanline  s.
    Function covering all steps which can be performed on one EPI.


    Parameters
    ----------
    epi : numpy.array [s,u]
        One gray-value epi.
    D: numpy.array [u].
        Already calculated disparities.
    Ce :  numpy.array [u]
        Edge confidence values for each EPI pixel.
    M : numpy.array [u] boolean.
       Mask which values to consider.
    s_hat : int
        The current scanline s to sample for.
    min_disp : int
        The minimal disparity to sample for.
    max_disp : int
        The maximal disparity to sample for.
    stepsize : float
        The stepsize used during the sampling procedure.
    Cd_t : float, optional
        The threshold which must be overcome for the depth confidence.
    NOISEFREE : boolean, optional
        Improve performance of noise-free EPIs.
    COARSEST : boolean, optional
        True means we are at the coarsest resolution.
    DEBUG : boolean, optional
        Enable plotting to visualize the sampling process
    s_hat_DEBUG : int, optional
        s_hat used for DEBUG output.
    DEBUG_dir : string, optional
        Directory used for DEBUG output.
    v_DEBUG : int, optional
        v used for DEBUG output.

    DB : numpy.array [u,2]
        The minimal and maximal disparity bounds to sample in between.
    M : numpy.array [u] boolean.
        Mask which values to consider.
    plots : numpy.array [d,s,u].
        If plotting was enabled epis to mark on the sampling process.
    DEBUG : boolean, optional
        Enable plotting to visualize the sampling process


    S : numpy.array [u,d]
        The calculated scores for each sampled radiance at each pixel u.

    R_bar : numpy.array [u,d].
        The updated convergence radiance for each EPI pixel.
    disp_range : numpy.array [d]
        The range of disparities used during the sampling.

    Returns
    -------
    D : numpy.array [u]
        The updated estimated disparities for each u at the current s-dimension.
    R_bar : numpy.array [u]
        The best updated convergence radiance for each EPI pixel.
    Md : numpy.array [u] of boolean.
        True means edge confidence is sufficient.
    Cd :  numpy.array [u]
        The calculated confidence values for each pixel.
    DB : numpy.array [u,2]
        The minimal and maximal disparity bounds to sample in between.
    S_max : numpy.array [u]
        The maximal score over all disparities per pixel.
    S_mean : numpy.array [u]
        The mean score over all disparities per pixel.
    S_argmax : numpy.array [u]
        The index of the maximal score.
    """

    # 2. Disparity bounds
    DB = find_disp_bounds(D, min_disp, max_disp,
                          stepsize)  # calcualte the current disparity bounds; ndarray[u]
    # 3. Radiance sampling(3)
    R, disp_range, plot = sample_radiance(epi, s_hat, min_disp, max_disp,
                                          stepsize, DB, M, DEBUG=DEBUG)
    # The sampled radiances; ndarray[u,d,s]. The range of disparities; ndarray[d]. The plots; ndarray[d,s,u]
    if DEBUG:  # We plot the sampling results but only in an exemplary fashion
        if s_hat == s_hat_DEBUG:
            imsave(os.path.join(DEBUG_dir,
                                'Sampling_v={v}_s={s}_disparity={d}.png'.format(
                                    v=v_DEBUG, s=s_hat_DEBUG, d=disp_range[0])),
                   plot[0])
            imsave(os.path.join(DEBUG_dir,
                                'Sampling_v={v}_s={s}_disparity={d}.png'.format(
                                    v=v_DEBUG, s=s_hat_DEBUG,
                                    d=disp_range[-1])), plot[-1])

    # 4. Score computation (4, 5)
    S, R_bar = score_computation(R, epi, s_hat, M, h=0.02, NOISEFREE=NOISEFREE)
    # The scores; ndarray[d,u]. The R_bar values; ndarray[d,u]


    # 5. Disparity confidence (7)
    if not COARSEST:  # only accept values with high depth confidence
        Cd, Md, S_max, S_mean = disparity_confidence(S, Ce, threshold=Cd_t)
    else:  # except for the lowest resolution
        Cd, Md , S_max, S_mean = disparity_confidence(S, Ce, threshold=-1)
        assert np.all(
            Md), 'Unvalide disparity confidence at coarsest resolution.'
    # The disparity confidence score; ndarray[u]. disparity confidence mask; ndarray[u].


    # 6. Disparity estiamte (6)
    D, R_bar, S_argmax = estimate_disp(D, S, M, R_bar, disp_range)
    # The best scored disparity estimate; ndarray[u]. The best scored R_bar; ndarray[u].

    return D, R_bar, Md, Cd, DB,  S_max, S_mean, S_argmax


def convert_process_epi(args):
    """
    Function to allow multiprocessing with multiple arguments.
    """

    return process_epi(*args)



#!/usr/bin/python
# -*- coding: utf8 -*-

"""
Disney Disparity.
:author: Manuel Tuschen
:date: 20.06.2016
:license: GPL3
"""

from __future__ import division, absolute_import, unicode_literals, print_function

import numpy as np
from numba import jit


def score_computation(R, epi, s_hat, M, h=0.02, NOISEFREE=False):
    """
    Calculates a score for each disparity estimate as in eq. 4:
    Afterwards for each EPI pixel u in the currently processed s-dimension each
    disparity is scored.

    Parameters
    ----------
    R : numpy.array [u,d,s]
       The set of sampled radiances. For each pixel u and disparity d the gray
       value for each scanline s is stored.
    epi : numpy.array [s,u]
       One gray-value epi.
    s_hat : int
       The current scanline s to sample for.
    M : numpy.array [u] boolean.
       Mask which values to consider.
    h : float
        A bandwidth parameter for the kernel function.
    NOISEFREE: boolean, optional
        Improve performance of noise-free EPIs.

    Returns
    -------
    S_norm : numpy.array [u,d]
        The calculated scores for each sampled radiance at each pixel u.
    R_bar : numpy.array [u,d].
        The updated convergence radiance for each EPI pixel.
    """

    u_dim = R.shape[0]
    d_dim = R.shape[1]
    s_dim = R.shape[2]
    # Check dimensions of input data
    assert epi.shape == (s_dim, u_dim,), 'Input EPI has wrong shape in function \'score_computation\'.'
    assert M.shape == (u_dim,), 'Input Mc has wrong shape in function \'score_computation\'.'
    assert s_hat >= 0 and s_hat < s_dim, 's_hat not in range 0 - s-dim in function \'sample\'.'

    # alculate initial R_bar. There must be one value for each disparity later.
    R_bar = np.tile(epi[s_hat].reshape((u_dim, 1, 1)), (1, d_dim, s_dim))
    assert R_bar.shape == (u_dim, d_dim, s_dim), 'Initial R_bar has wrong shape in fucntion \'score_computation\'.'
    assert not np.any(np.isnan(R_bar)), 'NaN detected in initial R_bar in function \'score_computation\'.'

    if not NOISEFREE:
        R_bar = calc_r_bar_iter(R_bar, R, M, h)
        assert R_bar.shape == (u_dim, d_dim, s_dim), 'Iterative R_bar has wrong shape in function \'score_computation\'.'
        assert not np.any(np.isnan(R_bar[:, :, s_hat][M])), 'NaN detected in final R_bar in function \'score_computation\'.'

    # calculate R - R_bar
    R_minus_R_bar = R - R_bar
    assert R_minus_R_bar.shape == (u_dim, d_dim,s_dim,), 'R - R_bar difference has wrong shape in function \'score_computation\'.'


    # compute the norm for the scoring function
    R_norm = np.sum(~np.isnan(R), axis=-1)  # count along s-axis ignoring NaNs
    all_nan = R_norm == 0
    R_norm += 1
    #R_norm[~all_nan] = 1.0 / R_norm[~all_nan] # Not working why ??
    R_norm = 1.0 / R_norm
    assert R_norm.shape == (u_dim, d_dim), 'R_norm output has wrong shape in fucntion \'score_computation\'.'

    # calculate the score
    S = np.sum(epanechnikov_kernel(R_minus_R_bar, M, h), axis=-1)
    S_norm = R_norm * S
    R_bar = R_bar[:, :, s_hat]

    # Let's see if our final results have reasonable meaning'
    assert S_norm.shape == (u_dim, d_dim), 'S_norm output has wrong shape in fucntion \'score_computation\'.'
    assert not np.any(S_norm[all_nan] != 0), 'Score not zero if all entries are NaN in fucntion \'score_computation\'.'
    assert not np.any(np.isnan(S_norm)), 'NaN detected in S_norm output in fucntion \'score_computation\'.'
    assert R_bar.shape == (u_dim, d_dim), 'R_bar output has wrong shape in fucntion \'score_computation\'.'
    return S_norm, R_bar


@jit(nopython=True)
def epanechnikov_kernel(R, M, h):
    """
    The epanechnikov kernel function as in eq. 4

    Parameters
    ----------
    R : numpy.array [u,d,s]
       The set of sampled radiances. For each pixel u and disparity d the gray
       value for each scanline s is stored.
    M : numpy.array [u] boolean.
       Mask which values to consider.
    h : float
        A bandwidth parameter for the kernel function.

    Returns
    -------
    kernel_val: numpy.array  [u,d,s].
        Float values between 0 and 1 which 1 being the highest score possible.
    """

    u_dim = R.shape[0]
    d_dim = R.shape[1]
    s_dim = R.shape[2]

    kernel_vals = np.zeros(R.shape, dtype=np.float64)

    for u in range(u_dim):
        if not M[u]:
            continue
        for d in range(d_dim):
            for s in range(s_dim):
                if not np.isnan(R[u, d, s]) and abs(R[u, d, s] / h) <= 1:  # can still be zero due to sampling outside of epi
                    kernel_vals[u, d, s] = 1 - (R[u, d, s] / h) ** 2

    return kernel_vals


@jit(nopython=True)
def calc_r_bar_iter(R_bar, R, M, h, n_iter=10):
    """
    Calculate the radiance mean according to eq. 5
    For each EPI pixel in the current s-dimension a new convergence radiance is
    sampled.

    Parameters
    ----------
    R_bar : numpy.array [u,d]
        The updated convergence radiance for each EPI pixel.
    R : numpy.array [u,d,s]
       The set of sampled radiances. For each pixel u and disparity d the gray
       value for each scanline s is stored.
    M : numpy.array [u] boolean.
        Mask which values to consider.
    h : float
        A bandwidth parameter for the kernel function.
    n_iter : int, optional
        The number of iterations for the convergence calculation.

    Returns
    -------
    R_bar : numpy.array [u,d]
        The updated convergence radiance for each EPI pixel.
    """

    u_dim = R.shape[0]
    d_dim = R.shape[1]
    s_dim = R.shape[2]

    n = 0
    while n < n_iter:
        R_minus_R_bar = R - R_bar
        assert R_minus_R_bar.shape == (u_dim, d_dim, s_dim), 'R - R_bar difference has wrong shape in function \'calc_r_bar_iter_loop\'.'

        kernel_vals = epanechnikov_kernel(R_minus_R_bar, M, h)
        assert kernel_vals.shape == (u_dim, d_dim, s_dim), 'kernel_vals has wrong shape in function \'calc_r_bar_iter\'.'
        assert not np.any(np.isnan(kernel_vals)), 'NaN detected in kernel output in function \'calc_r_bar_iter\'.'

        # sum_numerator = np.nansum(kernel_vals * R, axis=-1) # not supported yet
        sum_numerator = sum_along_s_axis(kernel_vals * R, M)  # resulting array will have NaNs as R !
        assert sum_numerator.shape == (u_dim, d_dim), 'sum_numerator has wrong shape in function \'calc_r_bar_iter\'.'

        # sum_denominator = np.nansum(kernel_vals, axis=-1) # not supported yet
        sum_denominator = sum_along_s_axis(kernel_vals, M)
        sum_denominator += 1  # add pseudocount to avoid division by zero
        assert sum_denominator.shape == (u_dim, d_dim), 'sum_denominator has wrong shape in function \'calc_r_bar_iter\'.'
        assert not np.any(np.isnan(sum_denominator[M])), 'NaN detected in sum_denominator output in function \'calc_r_bar_iter\'.'

        _R_bar = sum_numerator / sum_denominator
        assert _R_bar.shape == (u_dim, d_dim), 'R_bar has wrong shape in function \'calc_r_bar_iter\'.'

        #R_bar = np.tile(_R_bar[:,:,s_hat], (1, 1, s_dim)) # not supported yet
        R_bar = tile_R_bar(_R_bar, s_dim)
        assert R_bar.shape == (u_dim, d_dim,s_dim), 'R_bar output has wrong shape in function \'calc_r_bar_iter\'.'

        n += 1

    return R_bar




    ########################################################################
    #                                                                      #
    #                           Helper functions                           #
    #                                                                      #
    ########################################################################


# These functions are necessary as the standard way of doing it is not supported by numba right now


@jit(nopython=True)
def tile_R_bar(R_bar, s_dim):
    """
    Extend dimension of R_bar by an s-axis.

    Parameters
    ----------
    R_bar : numpy.array [u,d]
        The updated convergence radiance for each EPI pixel.
    s_dim : int
        The dimension of the s-axis.

    Returns
    -------
    R_bar_tiled : numpy.array [u,d, s]
        The extended R_bar array.
    """

    u_dim = R_bar.shape[0]
    d_dim = R_bar.shape[1]

    R_bar_tiled = np.zeros((u_dim, d_dim, s_dim), dtype=np.float64)

    for u in range(u_dim):
        for d in range(d_dim):
            for s in range(s_dim):
                R_bar_tiled[u, d, s] = R_bar[u, d]
    return R_bar_tiled


@jit(nopython=True)
def sum_along_s_axis(kernel_output, M):
    """
    Sum over the s dimension of the kernel values output.

    Parameters
    ----------
    kernel_output : numpy.array [u,d,s]
        Float values between 0 and 1 which 1 being the highest score possible.
    M : numpy.array [u] boolean
       Mask which values to consider.

    Returns
    -------
    kernel_sum : numpy.array [u,d]
        The sum of the kernel values over s dimension.
    """

    u_dim = kernel_output.shape[0]
    d_dim = kernel_output.shape[1]
    s_dim = kernel_output.shape[2]

    kernel_sum = np.full((u_dim, d_dim), np.nan, dtype=np.float64)
    for u in range(u_dim):
        if not M[u]:
            continue
        for d in range(d_dim):
            _sum = 0
            for s in range(s_dim):
                if not np.isnan(kernel_output[u, d, s]):
                    _sum += kernel_output[u, d, s]
            kernel_sum[u, d] = _sum
    return kernel_sum



#!/usr/bin/python
# -*- coding: utf8 -*-

"""
Disney Disparity.
:author: Manuel Tuschen
:date: 20.06.2016
:license: GPL3
"""

from __future__ import division, absolute_import, unicode_literals, print_function

import argparse
import os
import shutil
from itertools import repeat
from multiprocessing import Pool, cpu_count

import h5py
import matplotlib
import numpy as np
from progressbar import Bar, ETA, Percentage, ProgressBar

matplotlib.use('Agg')
import matplotlib.pyplot as plt
from scipy.ndimage.filters import median_filter
from skimage.io import imsave
from skimage.color import gray2rgb

from ._confidence import edge_confidence
from ._process_epi import process_epi, convert_process_epi
from ._disparity import bilateral_median, propagation
from ._scale import fine_to_course, course_to_fine
from .lf2epi import calculate_resolutions, downsample_lightfield, create_epis


class Disney():
    '''
    The class collecting all functionality and input parameters needed for the
    disparity calculation.
    '''

    def __init__(self, lightfield, lf_group, output_dir, working_dir='work_tmp/',
                 n_cpus=-1, r_start=None, s_hat=None, DEBUG=False):

        """
        Constructor

        Parameters
        ----------
        lightfield : string
            The filename of the lightfield including the directory.
        lf_group : string
            The group name inside the lightfield's hdf5 file.
        output_dir: string
            The directory to store the final results in.
        working_dir: string, optional
            A temporal directory to work in.
        n_cpus : int, optional
            The number of cpus to use. If -1 all cpus available will be used.
        r_start : tuble (), optional
            r_v and r_u to start with
        s_hat : int, optional
            If given, only this s-dimension will be calculated.
        DEBUG : Boolean, optional
            enable DEBUG output
        """

        self.DEBUG = DEBUG  # Plot intermediate results and other debugging output

        self.n_cpus = n_cpus
        if self.n_cpus == -1:
            self.n_cpus = int(cpu_count() * 0.9)

        self.output_dir = os.path.expanduser(output_dir)
        self.working_dir = os.path.expanduser(working_dir)
        if not os.path.exists(self.output_dir):
            os.makedirs(self.output_dir)
        if not os.path.exists(self.working_dir):
            os.makedirs(self.working_dir)

        # Attributes of .hdf5 files to load or store the data
        self.lf_group = lf_group
        self.light_field = h5py.File(os.path.expanduser(lightfield), 'r')
        self.epi_field = h5py.File(os.path.join(self.working_dir, 'epis.hdf5'),
                                   'a')
        self.disp_field = h5py.File(
            os.path.join(self.working_dir, 'disparities.hdf5'), 'a')

        if self.DEBUG:
            self.score_field = h5py.File(
                os.path.join(self.working_dir, 'scores.hdf5'), 'a')
            self.DB_field = h5py.File(
                os.path.join(self.working_dir, 'disparity_bounds.hdf5'), 'a')
            self.Ce_field = h5py.File(
                os.path.join(self.working_dir, 'edge_confidences.hdf5'), 'a')
            self.Cd_field = h5py.File(
                os.path.join(self.working_dir, 'disparity_confidences.hdf5'),
                'a')

        # Runtime attributes
        self.lf_res = self.light_field[
            self.lf_group].shape  # The resolution of the original lightfield (s,v,u)
        self.epi_res = None  # A ndarray [r]. The different EPI resolutions (v,u)

        self.r_start = self.lf_res[1:3]
        if self.DEBUG:
            self.r_start = self.lf_res[
                           1:3] if r_start is None else r_start  # If given start calculation by this resolution
        self.s_hat = s_hat  # If not None the only scanline to process

        self.initialize()
        print("All data loaded!")

    def __del__(self):
        """
        Cleanup.
        """
        self.light_field.close()
        self.epi_field.close()
        self.disp_field.close()
        if self.DEBUG:
            self.Ce_field.close()
            self.Cd_field.close()
            self.DB_field.close()
            self.score_field.close()
        else:
            shutil.rmtree(self.working_dir)


    def initialize(self):
        """
        Load data from files or initialize if not available.
        """

        # 2) We need all epi data:
        if not self.epi_field.get('epis', default=False):
            lf_file = self.light_field.filename
            self.light_field.close()
            epi_file = self.epi_field.filename
            self.epi_field.close()
            tmp_file = os.path.join(self.working_dir, 'tmp.hdf5')
            epi_res = calculate_resolutions(self.lf_res[1], self.lf_res[2])
            downsample_lightfield(lf_file, tmp_file, self.lf_group, epi_res)
            create_epis(tmp_file, epi_file, self.lf_group, dtype=np.float64,
                        RGB=False)
            os.remove(tmp_file)
            self.light_field = h5py.File(lf_file, 'r')
            self.epi_field = h5py.File(epi_file, 'r')
        self.epi_res = self.epi_field['epis'].attrs.get('resolutions')

        # 3) Check r-start
        if not np.all(np.less_equal(self.epi_res[-1], self.r_start)):
            raise IOError('r-start dimension not found')
        SCRATCH = False
        if np.array_equal(self.lf_res[1:3], self.r_start) or not self.DEBUG:
            SCRATCH = True

        # 4) We need all disparity data
        for res in self.epi_res:
            res_name = str(res[0]) + 'x' + str(res[1])
            if res[0] >= self.r_start[0] and res[1] >= self.r_start[
                1] and not SCRATCH:  # There should be data in file
                if not self.disp_field['disparities'].get(res_name,
                                                          default=False):
                    raise IOError(
                        'Diparity file object does not contain enough data sets.')

        if self.DEBUG:
            # 5) We need the initial disparitiy bounds for that resolution we start
            for res in self.epi_res:
                res_name = str(res[0]) + 'x' + str(res[1])
                if res[0] >= self.r_start[0] and res[1] >= self.r_start[
                    1] and not SCRATCH:  # There should be data in file
                    if not self.DB_field['disp_bounds'].get(res_name,
                                                            default=False):
                        raise IOError(
                            'Diparity file object does not contain enough data sets.')

            # 6) We need to initialize the edge confidence array
            for res in self.epi_res:
                res_name = str(res[0]) + 'x' + str(res[1])
                if res[0] > self.r_start[0] and res[1] > self.r_start[
                    1] and not SCRATCH:  # There must be data in file for at least all resolutions larger than r_start:
                    if not self.Ce_field['edge_conf'].get(res_name,
                                                          default=False):
                        raise IOError(
                            'Edge confidence file object does not contain enough data sets.')

            # 9) We need to initialize the disparity confidence array
            for res in self.epi_res:
                res_name = str(res[0]) + 'x' + str(res[1])
                if res[0] > self.r_start[0] and res[1] > self.r_start[
                    1] and not SCRATCH:  # There must be data in file for at least all resolutions larger than r_start:
                    if not self.Cd_field['disp_conf'].get(res_name,
                                                          default=False):
                        raise IOError(
                            'Disparity confidence file object does not contain enough data sets.')

    def getEpi(self, res, v=None, s=None, u=None):
        """
        Get the required epi data.

        Parameters
        ----------
        res : tuple
             The current v- and u-dimension.
        v : int, optional
            The only v-dimension to return.
        s : int, optional
            The only s-dimension to return.
        u : int, optional
            The only u-dimension to return.

        Returns
        -------
        : numpy.array [v,s,u]
            The sub(data) requested.

        """
        res_name = str(res[0]) + 'x' + str(res[1])
        data = self.epi_field['epis/' + res_name]

        if data is None:
            raise IOError('EPI data not found!')

        if v is None and s is None and u is None:
            return data[...]
        elif v is None and s is None:
            return data[:, :, u]
        elif v is None and u is None:
            return data[:, s]
        elif s is None and u is None:
            return data[v]
        elif v is None:
            return data[:, s, u]
        elif s is None:
            return data[v, :, u]
        elif u is None:
            return data[v, s]

    def getDisp(self, res, v=None, s=None, u=None, level=None):
        """
        Get the required required disparity data.

        Parameters
        ----------
        res : tuple
             The current v- and u-dimension.
        v : int, optional
            The only v-dimension to return.
        s : int, optional
            The only s-dimension to return.
        u : int, optional
            The only u-dimension to return.
        level : int, optional
            Determine which disparity data to get in DEBUG mode.
            0) after refinement,
            1) raw data,
            2) after bilateral median,
            3) after confidence selection

        Returns
        -------
        : numpy.array [v,s,u, (level)]
            The sub(data) requested.
        """
        grp = self.disp_field.require_group('disparities')
        res_name = str(res[0]) + 'x' + str(res[1])
        if self.DEBUG:
            grp.require_dataset(res_name,
                                        (res[0], self.lf_res[0], res[1], 4),
                                        dtype=np.float64, fillvalue=np.nan)
        else:
            grp.require_dataset(res_name,
                                        (res[0], self.lf_res[0], res[1]),
                                        dtype=np.float64, fillvalue=np.nan)
            level = None

        data = self.disp_field['disparities/' + res_name]

        if level is None:

            if v is None and s is None and u is None:
                return data[...]
            elif v is None and s is None:
                return data[:, :, u]
            elif v is None and u is None:
                return data[:, s]
            elif s is None and u is None:
                return data[v]
            elif v is None:
                return data[:, s, u]
            elif s is None:
                return data[v, :, u]
            elif u is None:
                return data[v, s]
        else:
            if v is None and s is None and u is None:
                return data[:, :, :, level]
            elif v is None and s is None:
                return data[:, :, u, level]
            elif v is None and u is None:
                return data[:, s, :, level]
            elif s is None and u is None:
                return data[v, :, :, level]
            elif v is None:
                return data[:, s, u, level]
            elif s is None:
                return data[v, :, u, level]
            elif u is None:
                return data[v, s, :, level]

    def getScore(self, res, v=None, s=None, u=None, level=None):
        """
        Get the required required score data.

        Parameters
        ----------
        res : tuple
             The current v- and u-dimension.
        v : int, optional
            The only v-dimension to return.
        s : int, optional
            The only s-dimension to return.
        u : int, optional
            The only u-dimension to return.
        level : int, optional
            Determine which score data to get in DEBUG mode.
            0) S_max,
            1) S_mean,
            2) S_argmax,

        Returns
        -------
        : numpy.array [v,s,u, (level)]
            The sub(data) requested.
        """
        grp = self.disp_field.require_group('scores')
        res_name = str(res[0]) + 'x' + str(res[1])

        grp.require_dataset(res_name,
                                    (res[0], self.lf_res[0], res[1],3),
                                    dtype=np.float64, fillvalue=np.nan)

        data = self.score_field['scores/' + res_name]

        if level is None:
            if v is None and s is None and u is None:
                return data[...]
            elif v is None and s is None:
                return data[:, :, u]
            elif v is None and u is None:
                return data[:, s]
            elif s is None and u is None:
                return data[v]
            elif v is None:
                return data[:, s, u]
            elif s is None:
                return data[v, :, u]
            elif u is None:
                return data[v, s]
        else:
            if v is None and s is None and u is None:
                return data[:, :, :, level]
            elif v is None and s is None:
                return data[:, :, u, level]
            elif v is None and u is None:
                return data[:, s, :, level]
            elif s is None and u is None:
                return data[v, :, :, level]
            elif v is None:
                return data[:, s, u, level]
            elif s is None:
                return data[v, :, u, level]
            elif u is None:
                return data[v, s, :, level]

    def getDBs(self, res, v=None, s=None, u=None):
        """
        Get the required required disparity bound data.

        Parameters
        ----------
        res : tuple
             The current v- and u-dimension.
        v : int, optional
            The only v-dimension to return.
        s : int, optional
            The only s-dimension to return.
        u : int, optional
            The only u-dimension to return.

        Returns
        -------
        : numpy.array [v,s,u]
            The sub(data) requested.
        """
        res_name = str(res[0]) + 'x' + str(res[1])
        data = self.DB_field['disp_bounds/' + res_name]

        if data is None:
            raise IOError('Disparity data not found!')

        if v is None and s is None and u is None:
            return data[...]
        elif v is None and s is None:
            return data[:, :, u]
        elif v is None and u is None:
            return data[:, s]
        elif s is None and u is None:
            return data[v]
        elif v is None:
            return data[:, s, u]
        elif s is None:
            return data[v, :, u]
        elif u is None:
            return data[v, s]

    def getCe(self, res, v=None, s=None, u=None):
        """
        Get the required required edge confidence data.

        Parameters
        ----------
        res : tuple
             The current v- and u-dimension.
        v : int, optional
            The only v-dimension to return.
        s : int, optional
            The only s-dimension to return.
        u : int, optional
            The only u-dimension to return.

        Returns
        -------
        : numpy.array [v,s,u]
            The sub(data) requested.
        """
        res_name = str(res[0]) + 'x' + str(res[1])
        grp = self.Ce_field['edge_conf']
        threshold = grp.attrs.get('threshold', default=None)
        data = self.Ce_field['edge_conf/' + res_name]

        if data is None:
            raise IOError('Edge confidence data not found!')

        if v is None and s is None and u is None:
            return data[...], threshold
        elif v is None and s is None:
            return data[:, :, u], threshold
        elif v is None and u is None:
            return data[:, s], threshold
        elif s is None and u is None:
            return data[v], threshold
        elif v is None:
            return data[:, s, u], threshold
        elif s is None:
            return data[v, :, u], threshold
        elif u is None:
            return data[v, s], threshold

    def getCd(self, res, v=None, s=None, u=None):
        """
        Get the required required disparity confidence data.

        Parameters
        ----------
        res : tuple
             The current v- and u-dimension.
        v : int, optional
            The only v-dimension to return.
        s : int, optional
            The only s-dimension to return.
        u : int, optional
            The only u-dimension to return.

        Returns
        -------
        : numpy.array [v,s,u]
            The sub(data) requested.
        """
        res_name = str(res[0]) + 'x' + str(res[1])
        grp = self.Cd_field['disp_conf']
        threshold = grp.attrs.get('threshold', default=None)

        data = self.Cd_field['disp_conf/' + res_name]

        if data is None:
            raise IOError('Disparity confidence data not found!')

        if v is None and s is None and u is None:
            return data[...], threshold
        elif v is None and s is None:
            return data[:, :, u], threshold
        elif v is None and u is None:
            return data[:, s, ...], threshold
        elif s is None and u is None:
            return data[v], threshold
        elif v is None:
            return data[:, s, u], threshold
        elif s is None:
            return data[v, :, u], threshold
        elif u is None:
            return data[v, s], threshold

    def saveDisp(self, data, res, v=None, s=None, u=None, level=None):
        """
        Save the selected disparity data.

        Parameters
        ----------
        data: numpy.array
            The data to save.
        res : tuple
             The current v- and u-dimension.
        v : int, optional
            The only v-dimension to return.
        s : int, optional
            The only s-dimension to return.
        u : int, optional
            The only u-dimension to return.
        level : int, optional
            Determine which disparity data to get in DEBUG mode.
            0) after refinement,
            1) raw data,
            2) after bilateral median,
            3) after confidence selection.
        """
        grp = self.disp_field.require_group('disparities')
        res_name = str(res[0]) + 'x' + str(res[1])
        if self.DEBUG:
            d_set = grp.require_dataset(res_name,
                                        (res[0], self.lf_res[0], res[1], 4),
                                        dtype=np.float64, fillvalue=np.nan)
        else:
            level = None
            d_set = grp.require_dataset(res_name,
                                        (res[0], self.lf_res[0], res[1]),
                                        dtype=np.float64, fillvalue=np.nan)

        if level is None:
            if v is None and s is None and u is None:
                d_set[...] = data
            elif v is None and s is None:
                d_set[:, :, u] = data
            elif v is None and u is None:
                d_set[:, s] = data
            elif s is None and u is None:
                d_set[v] = data
            elif v is None:
                d_set[:, s, u] = data
            elif s is None:
                d_set[v, :, u] = data
            elif u is None:
                d_set[v, s] = data
        else:
            if v is None and s is None and u is None:
                d_set[:, :, :, level] = data
            elif v is None and s is None:
                d_set[:, :, u, level] = data
            elif v is None and u is None:
                d_set[:, s, :, level] = data
            elif s is None and u is None:
                d_set[v, :, :, level] = data
            elif v is None:
                d_set[:, s, u, level] = data
            elif s is None:
                d_set[v, :, u, level] = data
            elif u is None:
                d_set[v, s, :, level] = data

    def saveScore(self, data, res, v=None, s=None, u=None, level=0):
        """
        Save the selected score data.

        Parameters
        ----------
        data: numpy.array
            The data to save.
        res : tuple
             The current v- and u-dimension.
        v : int, optional
            The only v-dimension to return.
        s : int, optional
            The only s-dimension to return.
        u : int, optional
            The only u-dimension to return.
        level : int, optional
            Determine which disparity data to get in DEBUG mode.
            0) S_max,
            1) S_mean,
            2) S_argmax,
        """
        grp = self.score_field.require_group('scores')
        res_name = str(res[0]) + 'x' + str(res[1])

        d_set = grp.require_dataset(res_name,
                                    (res[0], self.lf_res[0], res[1],3),
                                    dtype=np.float64, fillvalue=np.nan)

        if v is None and s is None and u is None:
            d_set[:, :, :, level] = data
        elif v is None and s is None:
            d_set[:, :, u, level] = data
        elif v is None and u is None:
            d_set[:, s, :, level] = data
        elif s is None and u is None:
            d_set[v, :, :, level] = data
        elif v is None:
            d_set[:, s, u, level] = data
        elif s is None:
            d_set[v, :, u, level] = data
        elif u is None:
            d_set[v, s, :, level] = data

    def saveDBs(self, data, res, v=None, s=None, u=None):
        """
        Save the selected disparity bound data.

        Parameters
        ----------
        data: numpy.array
            The data to save.
        res : tuple
             The current v- and u-dimension.
        v : int, optional
            The only v-dimension to return.
        s : int, optional
            The only s-dimension to return.
        u : int, optional
            The only u-dimension to return.
        """

        grp = self.DB_field.require_group('disp_bounds')
        res_name = str(res[0]) + 'x' + str(res[1])
        d_set = grp.require_dataset(res_name,
                                    (res[0], self.lf_res[0], res[1], 2),
                                    dtype=np.float64, fillvalue=0)

        if v is None and s is None and u is None:
            d_set[...] = data
        elif v is None and s is None:
            d_set[:, :, u] = data
        elif v is None and u is None:
            d_set[:, s] = data
        elif s is None and u is None:
            d_set[v] = data
        elif v is None:
            d_set[:, s, u] = data
        elif s is None:
            d_set[v, :, u] = data
        elif u is None:
            d_set[v, s] = data

    def saveCe(self, data, res, v=None, s=None, u=None, threshold=None):
        """
        Save the selected edge confidence data.

        Parameters
        ----------
        data: numpy.array
            The data to save.
        res : tuple
             The current v- and u-dimension.
        v : int, optional
            The only v-dimension to return.
        s : int, optional
            The only s-dimension to return.
        u : int, optional
            The only u-dimension to return.
        threshold : float, optional
            Confidence threshold for binary mask.
        """
        grp = self.Ce_field.require_group('edge_conf')
        if threshold is not None:
            grp.attrs.create('threshold', threshold)
        res_name = str(res[0]) + 'x' + str(res[1])
        d_set = grp.require_dataset(res_name, (res[0], self.lf_res[0], res[1]),
                                    dtype=np.float64, fillvalue=np.nan)

        if v is None and s is None and u is None:
            d_set[...] = data
        elif v is None and s is None:
            d_set[:, :, u] = data
        elif v is None and u is None:
            d_set[:, s] = data
        elif s is None and u is None:
            d_set[v] = data
        elif v is None:
            d_set[:, s, u] = data
        elif s is None:
            d_set[v, :, u] = data
        elif u is None:
            d_set[v, s] = data

    def saveCd(self, data, res, v=None, s=None, u=None, threshold=None):
        """
        Save the selected disparity confidence data.

        Parameters
        ----------
        data: numpy.array
            The data to save.
        res : tuple
             The current v- and u-dimension.
        v : int, optional
            The only v-dimension to return.
        s : int, optional
            The only s-dimension to return.
        u : int, optional
            The only u-dimension to return.
        threshold : float, optional
            Confidence threshold for binary mask.
        """

        grp = self.Cd_field.require_group('disp_conf')
        if threshold is not None:
            grp.attrs.create('threshold', threshold)
        res_name = str(res[0]) + 'x' + str(res[1])
        d_set = grp.require_dataset(res_name, (res[0], self.lf_res[0], res[1]),
                                    dtype=np.float64, fillvalue=np.nan)

        if v is None and s is None and u is None:
            d_set[...] = data
        elif v is None and s is None:
            d_set[:, :, u] = data
        elif v is None and u is None:
            d_set[:, s] = data
        elif s is None and u is None:
            d_set[v] = data
        elif v is None:
            d_set[:, s, u] = data
        elif s is None:
            d_set[v, :, u] = data
        elif u is None:
            d_set[v, s] = data

    def generate_s_hat_order(self):
        """
        Generate a list with each entry being the next s-entry to work with.

        Returns
        -------
         : list
         A ordered list with the s-dimensions to go through.

        """
        s = []
        if self.s_hat is not None:
            s.append(self.s_hat)  # We just use one scanline
        else:
            s.append(self.lf_res[0] // 2)
            for i in range(1, self.lf_res[0] // 2 + 1):
                s.append(self.lf_res[0] // 2 - i)  # The line under s_hat
                s.append(self.lf_res[0] // 2 + i)  # The line above s_hat
            if self.lf_res[
                0] % 2 == 0:  # For even numbers we have added one line too much
                s.pop()

        return s

    def calculateDisp(self, min_disp, max_disp, stepsize, Ce_t, Cd_t, S_t,
                      NOISEFREE=False):
        """
        The main method to calculate the disparity estimates.

        Parameters
        ----------
        min_disp : float
            The minimal disparity to sample for.
        max_disp : float
            The maximal disparity to sample for.
        stepsize : float
            The stepsize used during the sampling procedure.
        Ce_t : float
            The threshold for the edge confidence.
        Cd_t : float:
            The threshold for the disparity confidence.
        S_t : float
            The similarity threshold e.g. for the bilateral median filter.
        NOISEFREE : bool, optional
            True means not to iteratively smooth the mean radiance. This should
            only be enabled for noisy data.
        """
        s_list = self.generate_s_hat_order()  # generate list of s_hat lines to go through
        r_list = [r for r, res in enumerate(self.epi_res) if
                  res[0] <= self.r_start[0] and res[1] <= self.r_start[
                      1]]  # generate the indices of r we need to compute

        for r, res in enumerate(self.epi_res):  # adjust disparity bounds
            if res[0] > self.r_start[0] and res[1] > self.r_start[1]:
                u_scale = self.epi_res[r + 1][1] / self.epi_res[r][1]
                assert u_scale <= 1, 'The scaling factor of u-dimension is larger 1.'
                min_disp *= u_scale
                max_disp *= u_scale
                stepsize *= u_scale

        # Initialize a progess bar to track the calculation process
        max_val = long(0)
        for r in r_list:
            max_val += long(self.epi_res[r, 0] * len(s_list))
        widgets = ['Calculating disparities: ', Percentage(), ' ', Bar(), ' ',
                   ETA(), ' ']
        progress = ProgressBar(widgets=widgets, maxval=max_val).start()
        current_val = long(0)

        ###############################################################
        #                                                             #
        #                        Resolution level                     #
        #                                                             #
        ###############################################################
        for r in r_list:  # go through all epi resolutions; r is only the index of the resolution

            # create some variables and temporal arrays
            s_dim = self.lf_res[0]
            v_dim = self.epi_res[r, 0]
            u_dim = self.epi_res[r, 1]
            res = self.epi_res[r]

            if v_dim == self.lf_res[1] and u_dim == self.lf_res[2]:
                FINEST = True
            else:
                FINEST = False
            if v_dim == self.epi_res[-1, 0] and u_dim == self.epi_res[-1, 1]:
                COARSEST = True
            else:
                COARSEST = False

            # Create some directories
            sample_dir = os.path.join(self.output_dir,
                                      'Plots/Radiance/{v_dim}x{u_dim}/'.format(
                                          v_dim=v_dim, u_dim=u_dim))
            if self.DEBUG and not os.path.exists(sample_dir):
                os.makedirs(sample_dir)
            propagation_dir = os.path.join(self.output_dir,
                                           'Plots/Propagation/{v_dim}x{u_dim}/'.format(
                                               v_dim=v_dim, u_dim=u_dim))
            if self.DEBUG and not os.path.exists(propagation_dir):
                os.makedirs(propagation_dir)

            # Find number of jobs to start
            max_jobs = self.n_cpus
            min_task = int((v_dim * 0.3) // max_jobs) + 1

            n_jobs = min(max_jobs, v_dim // min_task)
            n_tasks = v_dim // n_jobs

            epis = self.getEpi(res)  # The epis; ndarray[v,s,u].
            Ds = self.getDisp(res, level=0)  # The disparities; ndarray[v,s,u].

            ###############################################################
            #                                                             #
            #                        scan line level                      #
            #                                                             #
            ###############################################################
            for s in s_list:  # scan through all lines in an epi; s is only the index of the line

                r_bars = np.zeros((v_dim, u_dim),
                                  dtype=np.float64)  # The scanline updated radiances of all epis; ndarray[v,u].
                Md = np.full((v_dim, u_dim), False,
                             dtype=np.bool)  # The scanline disparity confidences; ndarray[v,u].

                # 1. edge confidence (2):
                # The edge mask is needed on resolution level per scanline.
                # Thus we do the caclulation ones per resolution and store the
                # mask in memory for efficient access. 
                if not COARSEST:  # Only calculate EPI-pixels with high edge confidence
                    Ce, Me = edge_confidence(epis[:, s], window=9,
                                             threshold=Ce_t)
                else:  # except for the coarsest resolution
                    Ce, Me = edge_confidence(epis[:, s], window=9, threshold=-1)
                    assert np.all(
                        Me), 'Unvalide edge confidence at coarsest resolution.'

                if self.DEBUG:
                    self.saveCe(Ce, res, s=s, threshold=Ce_t)

                Mc = np.isnan(Ds[:, s])
                M = np.logical_and(Me,Mc)

                if n_jobs > 1:  # only then perform multiprocessing
                    pool = Pool(processes=n_jobs)

                    ###############################################################
                    #                                                             #
                    #                       epi level                             #
                    #                                                             #
                    ###############################################################
                    for v in range(0, v_dim,
                                   n_jobs * n_tasks):  # go through all epis; v is only the index of the epi

                        n = min(n_jobs * n_tasks, v_dim - v)
                        v_jobs = range(v, v + n)
                        s_hat_jobs = repeat(s, n)
                        Cd_t_jobs = repeat(Cd_t, n)
                        min_disp_jobs = repeat(min_disp, n)
                        max_disp_jobs = repeat(max_disp, n)
                        stepsize_jobs = repeat(stepsize, n)
                        NOISEFREE_jobs = repeat(NOISEFREE, n)
                        COARSEST_jobs = repeat(COARSEST, n)
                        DEBUG_jobs = repeat(self.DEBUG, n)
                        s_hat_DEBUG_jobs = repeat(s_list[0], n)
                        DEBUG_dir_jobs = repeat(sample_dir, n)
                        epi_jobs = iter([epis[v_] for v_ in v_jobs])
                        D_jobs = iter([Ds[v_, s] for v_ in v_jobs])
                        Ce_jobs = iter([Ce[v_] for v_ in v_jobs])
                        M_jobs = iter([M[v_] for v_ in v_jobs])


                        args = zip(epi_jobs, D_jobs, Ce_jobs,  M_jobs,
                                    s_hat_jobs, min_disp_jobs, max_disp_jobs,
                                    stepsize_jobs, Cd_t_jobs, NOISEFREE_jobs,
                                    COARSEST_jobs, DEBUG_jobs, s_hat_DEBUG_jobs,
                                    DEBUG_dir_jobs, v_jobs)

                        results_jobs = pool.imap(convert_process_epi, args,
                                                 chunksize=n_tasks)

                        for i, results in enumerate(results_jobs):
                            v_ = v_jobs[i]
                            Ds[v_, s] = results[0]
                            r_bars[v_] = results[1]
                            Md[v_] = results[2]

                            if self.DEBUG:
                                self.saveDisp(Ds[v_, s], res, v=v_, s=s,
                                              level=1)
                                self.saveCd(results[3], res, v=v_,
                                            threshold=Cd_t)
                                self.saveDBs(results[4], res, v=v_, s=s)
                                self.saveScore(results[5], res, v=v_, s=s, level=0)
                                self.saveScore(results[6], res, v=v_, s=s, level=1)
                                self.saveScore(results[7], res, v=v_, s=s, level=2)
                    pool.close()
                    pool.join()

                else:  # no multiprocessing
                    ###############################################################
                    #                                                             #
                    #                       epi level                             #
                    #                                                             #
                    ###############################################################
                    for v in range(v_dim):  # go through all epis; v is only the index of the epi

                        # 2. Disparity bounds
                        # 3. Radiance sampling(3)
                        # 4. Score computation (4, 5)
                        # 5. Disparity confidence (7)
                        # 6. Disparity estiamte (6)

                        Ds[v, s], r_bars[v], Md[v], Cd, DB, S_max, S_mean, S_argmax = process_epi(
                            epis[v], Ds[v, s], Ce[v], M[v], s, min_disp,
                            max_disp, stepsize, Cd_t, NOISEFREE, COARSEST,
                            DEBUG=self.DEBUG, s_hat_DEBUG=s_list[0],
                            DEBUG_dir=sample_dir, v_DEBUG=v)

                        if self.DEBUG:
                            self.saveDisp(Ds[v, s], res, v=v, s=s, level=1)
                            self.saveCd(Cd, res, v=v, threshold=Cd_t)
                            self.saveDBs(DB, res, v=v, s=s)
                            self.saveScore(S_max, res, v=v, s=s, level=0)
                            self.saveScore(S_mean, res, v=v, s=s, level=1)
                            self.saveScore(S_argmax, res, v=v, s=s, level=2)

                ###############################################################
                #                                                             #
                #                        scan line level                      #
                #                                                             #
                ###############################################################

                # 7. bilateral median filter
                if not COARSEST:  # using the threshold
                    Ds[:, s] = bilateral_median(Ds[:, s], epis[:, s],  M, Me,
                                                threshold=S_t)
                else:  # except for the lowest resolution, which is necessary not to introduce new NaNs
                    Ds[:, s] = bilateral_median(Ds[:, s], epis[:, s],  M, Me,
                                               threshold=float('inf'))
                if self.DEBUG:
                    self.saveDisp(Ds[:, s], res, s=s, level=2)

                M = np.logical_and(~Mc,~Md)
                Ds[:,s][M] = np.nan # remove unvalide values
                if self.DEBUG:
                    self.saveDisp(Ds[:, s], res, s=s, level=3)


                # 8. Propagate from scanline
                if self.s_hat is None:
                    Ds, epi_plot = propagation(Ds, epis, r_bars, s,
                                               threshold=S_t, DEBUG=False)
                    if self.DEBUG:
                        if s == s_list[0]:
                            imsave(os.path.join(propagation_dir,
                                                'Propagation_v={v}_s={s}.png'.format(
                                                    v=v_dim // 2, s=s)),
                                   epi_plot[:,s,:])

                if COARSEST:
                    assert not np.any(np.isnan(Ds[:, s]))
                self.saveDisp(Ds, res, level=0)

                current_val += long(v_dim)
                progress.update(current_val)


            ###############################################################
            #                                                             #
            #                        Resolution level                     #
            #                                                             #
            ###############################################################


            # One resolution is done, so we update which values need recomputation and save output

            if not COARSEST:
                res_next = self.epi_res[r + 1]
                u_scale = res_next[1] / res[1]
                assert u_scale <= 1, 'The scaling factor of u-dimension is larger 1.'
                min_disp *= u_scale
                max_disp *= u_scale
                stepsize *= u_scale

                Ds = fine_to_course(Ds, res_next)
                self.saveDisp(Ds, res_next, level=0)

        progress.finish()

        # at last we need to sample  up
        for r in range(len(self.epi_res) - 1, 0, -1):
            Ds = course_to_fine(self.getDisp(self.epi_res[r - 1], level=0),
                                self.getDisp(self.epi_res[r], level=0))
            self.saveDisp(Ds, self.epi_res[r - 1], level=0)

    def calculateMap(self):
        """
        Calculate the final disparity map for each s-dimension and save output
        into a hdf5 file container. In DEBUG mode lot's of furhter plot's are
        generated.
        """

        disp_dir = os.path.join(self.output_dir,'Plots/Disparity/{v_dim}x{u_dim}/'.format(v_dim=self.lf_res[1], u_dim=self.lf_res[2]))
        if not os.path.exists(disp_dir):
            os.makedirs(disp_dir)

        Disp_map_f = h5py.File(
            os.path.join(self.output_dir, 'disparity_map.hdf5'), 'w')
        Disp_map = Disp_map_f.require_dataset('disparity_map',shape=self.lf_res[0:3],dtype=np.float64)

        s_list = self.generate_s_hat_order()
        #Disp_map[:] = self.getDisp(self.epi_res[0], s=s level=0).swapaxes(0,1)#self.lf_res[0:3])

        for s in s_list:
            Disp_map[s] = median_filter(self.getDisp(self.epi_res[0], s=s, level=0), size=(3, 3))  # we apply a median filter to remove remaining speccles
            plt.plot()
            plt.imshow(Disp_map[s], cmap='gray', interpolation='None')
            plt.colorbar()
            plt.savefig(os.path.join(disp_dir, str(s) + '_final_disparityMap.png'))
            plt.close()

        if self.DEBUG:

            for r, res in enumerate(self.epi_res):
                s_dim = self.lf_res[0]
                v_dim = res[0]
                u_dim = res[1]

                lf_dir = os.path.join(self.output_dir,
                                      'Plots/lightfield/{v_dim}x{u_dim}/'.format(
                                          v_dim=v_dim, u_dim=u_dim))
                if not os.path.exists(lf_dir):
                    os.makedirs(lf_dir)

                disp_dir = os.path.join(self.output_dir,
                                        'Plots/Disparity/{v_dim}x{u_dim}/'.format(
                                            v_dim=v_dim, u_dim=u_dim))
                if not os.path.exists(disp_dir):
                    os.makedirs(disp_dir)

                score_dir = os.path.join(self.output_dir,
                                            'Plots/Scores/{v_dim}x{u_dim}/'.format(
                                                v_dim=v_dim, u_dim=u_dim))
                if not os.path.exists(score_dir):
                    os.makedirs(score_dir)

                bounds_dir = os.path.join(self.output_dir,
                                          'Plots/DisparityBounds/{v_dim}x{u_dim}/'.format(
                                              v_dim=v_dim, u_dim=u_dim))
                if not os.path.exists(bounds_dir):
                    os.makedirs(bounds_dir)

                dispConf_dir = os.path.join(self.output_dir,
                                            'Plots/DisparityConfidence/{v_dim}x{u_dim}/'.format(
                                                v_dim=v_dim, u_dim=u_dim))
                if not os.path.exists(dispConf_dir):
                    os.makedirs(dispConf_dir)

                edge_dir = os.path.join(self.output_dir,
                                        'Plots/Edges/{v_dim}x{u_dim}/'.format(
                                            v_dim=v_dim, u_dim=u_dim))
                if not os.path.exists(edge_dir):
                    os.makedirs(edge_dir)

                s = self.s_hat if self.s_hat is not None else s_list[
                    len(s_list) // 2]

                Lf_map = self.getEpi(self.epi_res[r], s=s)
                Disp_map = self.getDisp(self.epi_res[r], s=s)
                Score_map = self.getScore(self.epi_res[r], s=s)
                DB_map = self.getDBs(self.epi_res[r], s=s)
                Cd_map, Cd_t = self.getCd(self.epi_res[r], s=s)
                Md_map = Cd_map > Cd_t
                Ce_map, Ce_t = self.getCe(self.epi_res[r],s=s)
                Me_map = Ce_map > Ce_t


                plt.plot()
                plt.imsave(os.path.join(lf_dir, str(s) + '_ligthfield.png'),gray2rgb(Lf_map[...]))
                plt.close()

                plt.plot()
                plt.imshow(Disp_map[ :, :, 0], cmap='gray',interpolation='none')
                plt.colorbar()
                plt.savefig(os.path.join(disp_dir, str(s) + '_upsampled_disparityMap.png'))
                plt.close()

                plt.plot()
                plt.imshow(Disp_map[ :, :, 1], cmap='gray',interpolation='none')
                plt.colorbar()
                plt.savefig(os.path.join(disp_dir, str(s) + '_raw_disparityMap.png'))
                plt.close()

                plt.plot()
                plt.imshow(Disp_map[ :, :, 2], cmap='gray',interpolation='none')
                plt.colorbar()
                plt.savefig(os.path.join(disp_dir, str(s) + '_median_disparityMap.png'))
                plt.close()

                plt.plot()
                plt.imshow(Disp_map[ :, :, 3], cmap='gray',interpolation='none')
                plt.colorbar()
                plt.savefig(os.path.join(disp_dir, str(s) + '_confidendt_disparityMap.png'))
                plt.close()


                plt.plot()
                plt.imshow(Score_map[ :, :, 0], cmap='gray',interpolation='none')
                plt.colorbar()
                plt.savefig(os.path.join(score_dir, str(s) + '_S_max.png'))
                plt.close()

                plt.plot()
                plt.imshow(Score_map[ :, :, 1], cmap='gray',interpolation='none')
                plt.colorbar()
                plt.savefig(os.path.join(score_dir, str(s) + '_S_mean.png'))
                plt.close()

                plt.plot()
                plt.imshow(Score_map[ :, :, 2], cmap='gray',interpolation='none')
                plt.colorbar()
                plt.savefig(os.path.join(score_dir, str(s) + '_S_argmax.png'))
                plt.close()

                plt.plot()
                plt.imshow(DB_map[ :, :, 0], cmap='gray',interpolation='none')
                plt.colorbar()
                plt.savefig(os.path.join(bounds_dir, str(s) + '_lowerDisparityBoundsMap.png'))
                plt.close()

                plt.plot()
                plt.imshow(DB_map[ :, :, 1], cmap='gray',interpolation='none')
                plt.colorbar()
                plt.savefig(os.path.join(bounds_dir, str(s) + '_upperDisparityBoundsMap.png'))
                plt.close()

                plt.plot()
                plt.imshow(Cd_map[...], cmap='gray', interpolation='none')
                plt.colorbar()
                plt.savefig(os.path.join(dispConf_dir, str(s) + '_disparityConfidenceMap.png'))
                plt.close()

                plt.plot()
                plt.imshow(Md_map[...], cmap='gray', interpolation='none')
                plt.colorbar()
                plt.savefig(os.path.join(dispConf_dir, str(s) + '_disparityConfidenceMask.png'))
                plt.close()

                plt.plot()
                plt.imshow(Ce_map[...], cmap='gray', interpolation='none')
                plt.colorbar()
                plt.savefig(os.path.join(edge_dir, str(s) + '_egeConfidenceMap.png'))
                plt.close()

                plt.plot()
                plt.imshow(Me_map[...], cmap='gray', interpolation='none')
                plt.colorbar()
                plt.savefig(os.path.join(edge_dir,str(s) + '_egeConfidenceMask.png'))
                plt.close()


################################################################################
#                                                                              #
#                       Can be used as a command line tool                     #
#                                                                              #
################################################################################



if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='Calulate disparity fom a given lightfield.')

    parser.add_argument('lightfield',
                        help='The filename including the directory of the lightfield.')

    parser.add_argument('--hdf5_group', help='The group name inside hdf5 File.',
                        default='lightfield')
    parser.add_argument('--working_dir',
                        help='The directory used to store intermediate data.',
                        default='tmp/')
    parser.add_argument('--result_dir', help='The output directory.',
                        default='results/')

    parser.add_argument('--s_hat',
                        help='If given, only this s-dimension will be computed.',
                        type=int, default=None)
    parser.add_argument('--r_start',
                        help='The resolution to start with. Can only be used in DEBUG mode.',
                        nargs=2, type=int, default=None)

    parser.add_argument('--min_disp',
                        help='The minimal disparity to sample for.', type=float,
                        default=1)
    parser.add_argument('--max_disp',
                        help='The maximal disparity to sample for.', type=float,
                        default=20)
    parser.add_argument('--stepsize',
                        help='The disparity setp size during sampling .',
                        type=float, default=1)
    parser.add_argument('--Ce_t',
                        help='The threshold for the edge confidence measurement.',
                        type=float, default=0.02)
    parser.add_argument('--Cd_t',
                        help='The threshold for the depth confidence measurement.',
                        type=float, default=0.1)
    parser.add_argument('--S_t',
                        help='The threshold for the bilateral median filter measurement.',
                        type=float, default=0.1)

    parser.add_argument('--n_jobs', help='The number of threads to use.',
                        type=int, default=-1)

    parser.add_argument('-NOISEFREE',
                        help='Disable radiance update for lightfield without noise',
                        action='store_true')
    parser.add_argument('-DEBUG',
                        help='Enable plotting of intermediate results',
                        action='store_true')

    args = parser.parse_args()

    disney = Disney(args.lightfield, args.hdf5_group, args.result_dir,
                    working_dir=args.working_dir, n_cpus=args.n_jobs,
                    r_start=args.r_start, s_hat=args.s_hat, DEBUG=args.DEBUG)

    disney.calculateDisp(min_disp=args.min_disp, max_disp=args.max_disp,
                         stepsize=args.stepsize, Ce_t=args.Ce_t, Cd_t=args.Cd_t,
                         S_t=args.S_t, NOISEFREE=args.NOISEFREE)

    disney.calculateMap()



#!/usr/bin/python
# -*- coding: utf8 -*-

"""
Disney Disparity.
:author: Manuel Tuschen
:date: 20.06.2016
:license: GPL3
"""

from __future__ import division, absolute_import, unicode_literals, print_function

import os
import struct
import argparse

import numpy as np
import h5py
from easyScripting import multiLoading, prepareSaving


def dmap2hdf5(input_dir, output_file, output_dataset='disparities'):
    """
    Convert results from .dmap files into .hdf5 file.

    Parameters
    ----------
    input_dir : string
        The directory where the .dmap files are located.
    output_file: string
        The filename  (including the directory), of the output file.
    output_dataset : string, optional
        The new container name of the hdf5 file.
    """

    files = multiLoading(directory=input_dir, extension=".dmap")

    # prepare saving
    output_name = os.path.basename(output_file)
    output_dir = os.path.dirname(output_file)
    output_file = prepareSaving(output_name, path=output_dir, extension=".hdf5")

    f_out = h5py.File(output_file, 'w')


    # we need to know the dimensions to store
    with open(files[0], 'rb') as f:  # open file
        b = f.read()  # read bytes
        # print("#bytes: {b}".format(b=len(b)))
        w, h = struct.unpack('<2i', b[0:8])  # width and hight
        # print("width:  {w}, heigth: {h}".format(w=w, h=h))

    f_out.create_dataset(output_dataset, shape=(len(files), h, w),
                    dtype=np.float64)

    # load all data from file
    for f, file in enumerate(files):
        with open(file, 'rb') as fl:  # open file
            b = fl.read()  # read bytes
            # print("#bytes: {b}".format(b=len(b)))
            w, h = struct.unpack('<2i', b[0:8])  # width and hight
            # print("width:  {w}, heigth: {h}".format(w=w, h=h))
            data = struct.unpack('<' + str(w * h) + 'f', b[0:w * h * 4])  # data
            data = np.array(data, np.float32).reshape((w, h))
            # print("dimensions: {w}x{h}".format(w=data.shape[0], h=data.shape[1]))

        f_out[f] = data


    f_out.close()




################################################################################
#                                                                              #
#                       Can be used as a command line tool                     #
#                                                                              #
################################################################################



if __name__ == "__main__":

    parser = argparse.ArgumentParser(description='Convert results from .dmap files into .hdf5 file.')

    parser.add_argument('input_dir', help='The directory where the .dmap files are located.')
    parser.add_argument('output_file', help='TThe filename  (including the directory), of the output file.')
    parser.add_argument('--output_dataset', help='The container name in the hdf5 file.', default='disparities')

    args = parser.parse_args()

    dmap2hdf5(args.input_dir, args.output_file, output_dataset=args.output_dataset)


#!/usr/bin/python
# -*- coding: utf8 -*-

"""
Disney Disparity.
:author: Manuel Tuschen
:date: 20.06.2016
:license: GPL3
"""

from __future__ import division, absolute_import, unicode_literals, print_function

import os
import argparse

import numpy as np
import h5py
from easyScripting import prepareLoading, prepareSaving



def clif2lf(clif_file, lf_file, clif_group, lf_group="lightfiled" ):
    """
    Convert a standard .clif file to an .hdf5 lightfield file.

    Parameters
    ----------
    clif_file : string
        The .clif filename including the directory.
    lf_file : string
        The filename (including the directory) of the output .hdf5 lightfield.
    clif_group : string
        The container name inside the .clif file.
    lf_group : string, optional
        The container name inside the .hdf5 file for the lightfield.
    """

    # Initialze the hdf5 file objects
    fname_in = os.path.basename(clif_file)
    dir_in = os.path.dirname(clif_file)
    clif_file = prepareLoading(fname_in, path=dir_in)

    fname_out = os.path.basename(lf_file)
    dir_out = os.path.dirname(lf_file)
    lf_file = prepareSaving(fname_out, path=dir_out, extension=".hdf5")

    data = h5py.File(clif_file, 'r')[clif_group]
    data = np.swapaxes(data, 1, 3 )
    data = np.swapaxes(data, 1, 2)

    f_out = h5py.File(lf_file, 'w')
    f_out.create_dataset(lf_group, data=data)
    f_out.close()




if __name__ == "__main__":

    parser = argparse.ArgumentParser(description='Convert lightfiled of .clif files to the .hdf5 file format.')
    parser.add_argument('clif_file', help='The filename including the directory of the .clif file.')
    parser.add_argument('lf_file', help='The filename including the directory of the .ddf5 file.')
    parser.add_argument('clif_group', help='The group name inside .clif File.')
    parser.add_argument('--lf_group', help='The group name inside .hdf5 File.', default='lightfield')

    args = parser.parse_args()
    clif2lf(args.lightfiled, args.lf_file, args.clif_group, lf_group=args.lf_group)




#!/usr/bin/python
# -*- coding: utf8 -*-

"""
DisneyDisp.
:author: Manuel Tuschen
:date: 04.07.2016
:license: GPL3
"""


from __future__ import division, absolute_import, unicode_literals, print_function

from .imgs2lf import *
from .lf2epi import *
from .clif2lf import *
from .dmap2hdf5 import *

from ._confidence import *
from ._radiance import *
from ._score import *
from ._disparity import *
from ._process_epi import *
from ._scale import *
from .disney import *




#!/usr/bin/python
# -*- coding: utf8 -*-

"""
Disney Disparity.
:author: Manuel Tuschen
:date: 20.06.2016
:license: GPL3
"""

from __future__ import division, absolute_import, unicode_literals, print_function

from math import floor, ceil

import numpy as np
from numba import jit


@jit(nopython=True)
def find_disp_bounds(Disps, min_disp, max_disp, stepsize):
    """
    Calculate the disparity bounds for the next iteration

    Parameters
    ----------
    Disps : numpy.array [u]
        Reliable disparity estimates. All other values must be NaN.
    min_disp : int
        The minimal disparity to sample for.
    max_disp : int
        The maximal disparity to sample for.
    stepsize : float
        The stepsize used during the sampling procedure.

    Returns
    -------
    DB : numpy.array [u,2]
        The estimated disparity bounds for each u at line s in each EPI.
    """

    u_dim = Disps.shape[0]

    DB = np.zeros((u_dim, 2), dtype=np.float64)
    DB[:, 0] = min_disp
    DB[:, 1] = max_disp

    if np.all(np.isnan(Disps)):
        # if there are no disparity estiamtes yet
        return DB
    else:

        # counter = np.sum(~np.isnan(Disps)) # Does not work in numba yet
        counter = 0  # count how many valid estimates there are already
        for u in range(u_dim):
            if not np.isnan(Disps[u]):
                counter += 1
        ind = np.zeros((counter), dtype=np.uint)
        i = 0
        for u in range(u_dim):  # we seek for non nan indeces
            if not np.isnan(Disps[u]):
                ind[i] = u
                i += 1

        if len(ind) > 0:
            for i in range(len(ind)):
                DB[ind[i], 0] = Disps[
                    ind[i]]  # reliable estimates keep their bound
                DB[ind[i], 1] = Disps[ind[i]]
                assert DB[ind[i], 0] == DB[ind[i], 1], 'Depth bounds for reliable estimates are nor equal in function \'disp_bounds\'.'

        if len(ind) > 1:
            i = 0
            first = ind[0]
            while i < len(
                    ind) - 1:  # we assign boarders only if pixels are in between two valid estimates
                i += 1
                second = ind[i]

                min_DB = min(Disps[first], Disps[second]) - stepsize
                max_DB = max(Disps[first], Disps[second]) + stepsize
                for j in range(int(first + 1), int(second)):
                    DB[j, 0] = min_DB
                    DB[j, 1] = max_DB
                    assert DB[j, 0] < DB[j, 1], 'First depth bound not smaller than second in function \'disp_bounds\'.'
                first = second

        return DB


def sample_radiance(epi, s_hat, min_disp, max_disp, stepsize, DB, M, DEBUG=False):
    """
    Sample radiances according to eq. 3:
    R(u,d) = { E( u+(s_hat - s)d ,s ) | s = 1, ..., n }

    This function works on one scanline s per EPI. For each EPI-pixel u in that
    line, radiances are sampled for all disparities d. If disparity bounds are
    given, only disparities in between the bounds or equal are taken into
    account.

    Parameters
    ----------
    epi : numpy.array [s,u]
        One gray-value epi.
    s_hat : int
        The current scanline s to sample for.
    min_disp : int
        The minimal disparity to sample for.
    max_disp : int
        The maximal disparity to sample for.
    stepsize : float
        The stepsize used during the sampling procedure.
    DB : numpy.array [u,2]
        The minimal and maximal disparity bounds to sample in between.
    M : numpy.array [u] boolean.
        Mask which values to consider.
    DEBUG : boolean, optional
        Enable plotting to visualize the sampling process

    Returns
    -------
    R : numpy.array [u,d,s]
        The set of sampled radiances. For each pixel u and disparity d the gray
        value for each scanline s is stored.
    disp_range : numpy.array [d].
        The range of disparities used during the sampling.
    plots : ndarray [d,s,u] or None.
        If plotting was enabled epis with markings of the sampling process.
    """

    # Calculate all radiances to sample for
    n_disp = int((max_disp - min_disp) / stepsize) + 1
    disp_range = np.linspace(min_disp, max_disp, n_disp)

    u_dim = epi.shape[1]
    d_dim = n_disp
    s_dim = epi.shape[0]
    # Check dimensions of input data
    assert epi.shape == (s_dim, u_dim,), 'Input EPI has wrong shape in function \'sample_radiance\'.'
    assert DB.shape == (u_dim, 2), 'Input DB has wrong shape in function \'sample_radiance\'.'
    assert M.shape == (u_dim,), 'Input M has wrong shape in function \'sample_radiance\'.'
    assert d_dim > 0, 'There are no disparities to samle for in function \'sample_radiance\'.'
    assert min_disp < max_disp, 'minmimal disparity is not smaller than maximal disparity in function \'sample_radiance\'.'
    assert s_hat >= 0 and s_hat < s_dim, 's_hat {s_hat} not in range 0 - s-dim in function \'sample_radiance\'.'

    if DEBUG:
        plots = np.tile(epi.reshape((1, s_dim, u_dim,)), (n_disp, 1, 1,))
    else:
        plots = np.zeros((n_disp, s_dim, u_dim,), dtype=np.uint8)

    R, plots = sample_radiance_inner(epi, s_hat, disp_range, DB, M, plots, DEBUG=DEBUG)

    # Let's see if our results have reasonable meaning'
    assert R.shape == (u_dim, d_dim, s_dim,), 'Output R has wrong shape in function \'sample_radiance\'.'
    assert plots.shape == (d_dim, s_dim, u_dim,), 'Output R has wrong shape in function \'sample_radiance\'.'

    return R, disp_range, plots


@jit(nopython=True)
def sample_radiance_inner(epi, s_hat, disp_range, DB, M, plots, DEBUG=False):
    """
    A fast inner function for sampling the radinaces according to eq. 3:

    Parameters
    ----------
    epi : numpy.array [s,u]
        One gray-value epi.
    s_hat : int
        The current scanline s to sample for.
    min_disp : int
        The minimal disparity to sample for.
    max_disp : int
        The maximal disparity to sample for.
    stepsize : float
        The stepsize used during the sampling procedure.
    DB : numpy.array [u,2]
        The minimal and maximal disparity bounds to sample in between.
    M : numpy.array [u] boolean.
        Mask which values to consider.
    plots : numpy.array [d,s,u].
        If plotting was enabled epis to mark on the sampling process.
    DEBUG : boolean, optional
        Enable plotting to visualize the sampling process

    Returns
    -------
    R : numpy.array [u,d,s]
        The set of sampled radiances. For each pixel u and disparity d the gray
        value for each scanline s is stored.
    plots : ndarray [d,s,u] or None.
        If plotting was enabled epis with markings of the sampling process.
    """

    u_dim = epi.shape[1]
    d_dim = len(disp_range)
    s_dim = epi.shape[0]

    R = np.full((u_dim, d_dim, s_dim,), np.nan, dtype=np.float64)

    for u in range(u_dim):
        if not M[u]:
            continue

        for d in range(d_dim):
            if disp_range[d] < DB[u, 0]:
                continue
            if disp_range[d] > DB[u, 1]:
                break # for some disparities there will be NaNs

            for s in range(s_dim):
                x = u + (s_hat - s) * disp_range[d]
                x0 = int(floor(x))
                x1 = int(ceil(x))

                if x0 >= 0 and x1 < u_dim:

                    if x0 == x1:  # if we match an epi pixel
                        a = 1
                        b = 0
                    else:  # otherwise interpolate
                        a = (x1 - x) / (x1 - x0)
                        b = (x - x0) / (x1 - x0)
                    R[u, d, s] = ((a * epi[s, x0]) + (b * epi[s, x1]))

                if DEBUG:
                    plots[d, s, x0] = 1
                    plots[d, s, x1] = 1

    return R, plots



#!/usr/bin/python
# -*- coding: utf8 -*-

"""
Disney Disparity.
:author: Manuel Tuschen
:date: 20.06.2016
:license: GPL3
"""

from __future__ import division, absolute_import, unicode_literals, print_function

import os
import math
import argparse

import numpy as np
import h5py
from progressbar import Bar, ETA, Percentage, ProgressBar

from skimage.util import img_as_float, img_as_ubyte, img_as_uint
from skimage.filters import gaussian
from skimage.transform import resize
from skimage.color import rgb2gray, gray2rgb
from easyScripting import prepareLoading, prepareSaving




def calculate_resolutions(r0_v, r0_u, red_fac=2, min_res=11):
    """
    Create a list of downsampled resolutions (v,u) by reducing the initial
    resolution by a constant factor for each dimension up to a minimal
    resolution.

    Parameters
    ----------
    r0_v : uint16
        The initial resolution in v-dimension.
    r0_u : uint16
        The initial resolution in u-dimension.
    red_fac : uint , optional
        The reduction factor used for down sampling. Default is to halve the
        resolution each time.
    min_res: uint16.
        The minimal resolution to sample to. The program will stop when either u
        or v reach min_res.

    Returns
    -------
    array_like
        Each entry is a tuple (u,v) of resolutions.
    """

    r_all = [[r0_v, r0_u]] # initialize the output list

    v = math.ceil(float(r0_v) / float(red_fac))
    u = math.ceil(float(r0_u) / float(red_fac))

    while v > min_res and u > min_res:
        r_all.append([v,u])
        u = math.ceil(float(u) / float(red_fac))
        v = math.ceil(float(v) / float(red_fac))

    return np.array(r_all, dtype=np.uint16)




def downsample_lightfield(lf_in, lf_out, hdf5_group, r_all):
    """
    Reduces the dimension of the input lightfield to the values given.
    Results are stored in a new hdf5 file.

    Parameters
    ----------
    lf_in : string
        The input hdf5 filename (including the directory) of the lightfield.
    lf_out : string
        The output hdf5 filename (including the directory) of the lightfield
        in all resolutions.
    hdf5_group: string
        The container name inside the hdf5 file for the lightfield. The same
        name will be used for the new file.
    r_all: array_like
        All resolutions to create. Each entry is a tuple (u,v) of resolutions.
    """

    # Initialze the hdf5 file objects
    fname_in = os.path.basename(lf_in)
    dir_in = os.path.dirname(lf_in)
    lf_in = prepareLoading(fname_in, path=dir_in)

    fname_out = os.path.basename(lf_out)
    dir_out = os.path.dirname(lf_out)
    lf_out = prepareSaving(fname_out, path=dir_out, extension=".hdf5")

    lf_in = h5py.File(lf_in, 'r')
    lf_out = h5py.File(lf_out, 'w')

    data_in = lf_in[hdf5_group]
    
    
    # We need to store all resolutions
    grp_out = lf_out.create_group(hdf5_group)
    grp_out.attrs.create('resolutions', r_all)
    
    
    # Initialize a progress bar to follow the downsampling
    widgets = ['Downsaple lightfields: ', Percentage(), ' ', Bar(),' ', ETA(), ' ']
    progress = ProgressBar(widgets=widgets, maxval=r_all.shape[0]).start()

    for r,res in enumerate(r_all):
        
        data_out = grp_out.create_dataset(str(res[0]) + 'x' + str(res[1]), shape=(data_in.shape[0], res[0], res[1], data_in.shape[3]), dtype=np.float64)
            
        if r == 0: # at lowest resolution we take the original image
            for s in range(data_in.shape[0]):
                data_out[s] = img_as_float(data_in[s])
        else: # we smooth the imput data
            data_prior = grp_out[str(r_all[r-1][0]) + 'x' + str(r_all[r-1][1])]
            for s in range(data_in.shape[0]):
                data_smoothed = img_as_float(gaussian(data_prior[s], sigma=math.sqrt(0.5), multichannel=True))
                data_out[s] = resize(data_smoothed, (res[0], res[1]))
        
        progress.update(r)
    progress.finish()

    # Cleanup
    lf_in.close()
    lf_out.close()




def create_epis(lf_in, epi_out, hdf5_group_in, hdf5_group_out="epis",  dtype=np.float64, RGB=True):
    """
    Create epis for all resolutions given by the input lightfield.


    Parameters
    ----------
    lf_in : string
        The input hdf5 filename (including the directory) of the lightfield.
    epi_out : string
        The output hdf5 filename (including the directory) of the lightfield
        in all resolutions.
    hdf5_group_in: string
        The container name inside the hdf5 file for the lightfield. The same
        name will be used for the new file.
    hdf5_group_out: string, optional
        The container name inside the hdf5 file for the epis.
    dtype : numpy.dtype, optional
        The new data type for the epis. Must be either
        np.float64, np.uint8 or np.uint16.
    RGB : bool, optional
        If True, the output epis will be converted to RGB (default).
        Otherwise gray type images are stored.
    """

    # Initialze the hdf5 file objects
    fname_in = os.path.basename(lf_in)
    dir_in = os.path.dirname(lf_in)
    lf_in = prepareLoading(fname_in, path=dir_in)

    fname_out = os.path.basename(epi_out)
    dir_out = os.path.dirname(epi_out)
    epi_out = prepareSaving(fname_out, path=dir_out, extension=".hdf5")

    # Initialze the hdf5 file objects
    lf_in = h5py.File(lf_in, 'r')
    epi_out = h5py.File(epi_out, 'w')

    r_all = lf_in[hdf5_group_in].attrs.get('resolutions')[...]

    epi_grp = epi_out.create_group(hdf5_group_out)
    epi_grp.attrs.create('resolutions', r_all)



    # Initialize a progress bar to follow the conversion
    widgets = ['Create EPIs: ', Percentage(), ' ', Bar(),' ', ETA(), ' ']
    progress = ProgressBar(widgets=widgets, maxval=r_all.shape[0]).start()
    for r,res in enumerate(r_all):
        progress.update(r)

        set_name = str(res[0]) + 'x' + str(res[1])
        lf_data = lf_in[hdf5_group_in + '/' + set_name]

        # Do we shall take RGB or gray images?
        channels = lf_data.shape[-1]
        if channels != 3 and channels != 1:
            raise TypeError('The lightfield is neither gray nor RGB!')
        if RGB:
            epi_data = epi_grp.create_dataset(set_name, shape=(res[0],lf_data.shape[0], res[1],3), dtype=dtype)
        else:
            epi_data = epi_grp.create_dataset(set_name, shape=(res[0],lf_data.shape[0], res[1],), dtype=dtype)


        for v in range(res[0]):

            if dtype is np.float64:
                if RGB and channels == 1:
                    epi_data[v] = img_as_float(gray2rgb(lf_data[:,v,])).reshape(epi_data[v].shape)
                elif not RGB and channels == 3:
                    epi_data[v] = img_as_float(rgb2gray(lf_data[:,v,])).reshape(epi_data[v].shape)
                else:
                    epi_data[v] = img_as_float(lf_data[:,v,...]).reshape(epi_data[v].shape)
            elif dtype is np.uint16:
                if RGB and channels == 1:
                    epi_data[v] = img_as_uint(gray2rgb(lf_data[:,v,])).reshape(epi_data[v].shape)
                elif not RGB and channels == 3:
                    epi_data[v] = img_as_uint(rgb2gray(lf_data[:,v,])).reshape(epi_data[v].shape)
                else:
                    epi_data[v] = img_as_uint(lf_data[:,v,...]).reshape(epi_data[v].shape)
            elif dtype is np.uint8:
                if RGB and channels == 1:
                    epi_data[v] = img_as_ubyte(gray2rgb(lf_data[:,v,...])).reshape(epi_data[v].shape)
                elif not RGB and channels == 3:
                    epi_data[v] = img_as_ubyte(rgb2gray(lf_data[:,v,...])).reshape(epi_data[v].shape)
                else:
                    epi_data[v] = img_as_ubyte(lf_data[:,v,...]).reshape(epi_data[v].shape)
            else:
                raise TypeError('Given dtype not supported.')

    progress.finish()


    # Cleanup
    lf_in.close()
    epi_out.close()




################################################################################
#                                                                              #
#                       Can be used as a command line tool                     #
#                                                                              #
################################################################################



if __name__ == "__main__":

    parser = argparse.ArgumentParser(description='Extract EPIs from a given lightfield and store them in a .hdf5 file.')

    parser.add_argument('lightfiled', help='The filename including the directory of the lightfield.')

    parser.add_argument('epis', help='The filename including the directory to save the EPI .hdf5 file.')
    parser.add_argument('--hdf5_group_lf', help='The group name inside hdf5 File.', default='lightfield')
    parser.add_argument('--hdf5_group_epi', help='The group name inside hdf5 File.', default='lightfield')
    parser.add_argument('--dtype', help='The dtype used to store the lightfild data.', choices=[np.float64, np.uint8, np.uint16], default=np.float64)
    parser.add_argument('-RGB', help='Flag to determine if resulting lightfield should consists of RGB values.', action='store_true')


    args = parser.parse_args()

    lf = h5py.File(args.lightfiled, 'r')
    r_all = calculate_resolutions(lf[args.h5path].shape[1], lf[args.h5path].shape[2])

    downsample_lightfield(args.lightfiled, 'tmp.hdf5', args.hdf5_group_lf, r_all)
    create_epis('tmp.hdf5', args.epi, args.hdf5_group_lf, hdf5_group_out=args.hdf5_group_epi, dtype=args.dtype, RGB=args.RGB)
    os.remove('tmp.hdf5')














#!/usr/bin/python
# -*- coding: utf8 -*-

"""
Disney Disparity.
:author: Manuel Tuschen
:date: 20.06.2016
:license: GPL3
"""

from __future__ import division, absolute_import, unicode_literals, print_function

import warnings
from math import ceil, floor

import numpy as np
from numba import jit


def estimate_disp(D, S, M, R_bar, disp_range):
    """
    Depth estimates according to eq. 6 for one scanline  s.
    Estimates are only valid, if the values satisfy the edge and depth
    confidence and recomputation is required.

    Parameters
    ----------
    D : numpy.array [u].
        Already calculated disparities.
    S : numpy.array [u,d]
        The calculated scores for each sampled radiance at each pixel u.
    M : numpy.array [u] boolean.
       Mask which values to consider.
    R_bar : numpy.array [u,d].
        The updated convergence radiance for each EPI pixel.
    disp_range : numpy.array [d].
        The range of disparities used during the sampling.

    Returns
    -------
    D : numpy.array [u]
        The updated estimated disparities for each u at the current s-dimension.
    R_bar_best : numpy.array [u].
        The best updated convergence radiance for each EPI pixel.
    S_argmax : numpy.array [u]
        The index of the maximal score.
    """

    u_dim = S.shape[0]
    d_dim = S.shape[1]
    # Check dimensions of input data
    assert D.shape == (u_dim,), 'Input D has wrong number of dimensions in function \'estimate_disp\'.'
    assert S.shape == (u_dim, d_dim), 'Input S has wrong number of dimensions in function \'estimate_disp\'.'
    assert M.shape == (u_dim,), 'Input Mc has wrong number of dimensions in function \'estimate_disp\'.'
    assert R_bar.shape == (u_dim,d_dim,), 'Input R_bar has wrong number of dimensions in function \'estimate_disp\'.'
    assert disp_range.shape == (d_dim,), 'Input disp has wrong number of dimensions in function \'estimate_disp\'.'
    assert not np.any(np.isnan(S)), 'NaN detected in Score input in fucntion \'estimate_disp\'.'

    # We need to find the disparity of maximal score
    S_argmax = np.argmax(S, axis=-1)
    D_best = disp_range[S_argmax]
    R_bar_best = R_bar[np.arange(u_dim), S_argmax]

    # Let's see if our results have reasonable meaning
    assert S_argmax.shape == (u_dim,), 'S_argmax has wrong shape in funcion \'estimate_disp\'.'
    assert D_best.shape == (u_dim,), 'D_best output has wrong shape in funcion \'estimate_disp\'.'
    assert not np.any(np.isnan(D_best)), 'NaN detected in D_best output in fucntion \'estimate_disp\'.'
    assert R_bar_best.shape == (u_dim,), 'R_bar_best output has wrong shape in function \'estimate_disp\'.'

    D[M] = D_best[M]

    return D, R_bar_best, S_argmax


def bilateral_median(Ds, epis, M, Me, window=11, threshold=0.1):
    """
    Smooth the calculated disparities according to eq. 8
    It is assumed that epis and edge_masks are symmetrically distributed around
    the current epi/mask.

    Parameters
    ----------
    Ds : numpy.array [v,u].
        The set of already calculated disparities for each EPI at one scanline.
    epi : numpy.array [v,u]
        Set of all gray-value epis for one scanline.
    M: numpy.array [u] boolean.
       Mask which values to consider.
    Me : numpy.array [v,u] of boolean.
        True means an edge was discovered for at that pixel.
    window : int, optional
        The window size for the bilateral median filter. As the window should
        be centered symetrically around the pixel to process the value shpuld
        be odd. For even numbers the next higher odd number is chosen.
    threshold : float, optional
        The threshold for the filter to determine if two EPI pixels are
        regarded as similar.

    Returns
    -------
    Ds: numpy.array [v,u].
        The smoothed disparities at the current s-dimension.
    """

    v_dim = epis.shape[0]
    u_dim = epis.shape[1]

    # Make window size odd
    if window % 2 == 0:
        warnings.warn(
            'window should be an odd number in function \'bilateral_median\'.')
        window += 1

    # Check dimensions of input data
    assert epis.shape == (v_dim, u_dim,), 'Input epis has wrong number of dimensions in function \'bilateral_median\'.'
    assert Ds.shape == (v_dim, u_dim,), 'Input Ds has wrong number of dimensions in function \'bilateral_median\'.'
    assert Me.shape == (v_dim,u_dim,), 'Input Mes has wrong number of dimensions in function \'bilateral_median\'.'
    assert M.shape == (v_dim, u_dim,), 'Input Mcs has wrong number of dimensions in function \'bilateral_median\'.'

    # The epis must be extended in u and v-dimension to avoid a boarder problem
    # pad epis in u-dimension
    epis_padded = np.pad(epis, ((0, 0), (int(window / 2), int(window / 2))),
                         'edge')
    Me_padded = np.pad(Me, ((0, 0), (int(window / 2), int(window / 2))), 'edge')
    Ds_padded = np.pad(Ds, ((0, 0), (int(window / 2), int(window / 2))), 'edge')

    # pad epis in v-dimension
    epis_padded = np.pad(epis_padded,
                         ((int(window / 2), int(window / 2)), (0, 0)), 'edge')
    Me_padded = np.pad(Me_padded, ((int(window / 2), int(window / 2)), (0, 0)),
                       'edge')
    Ds_padded = np.pad(Ds_padded, ((int(window / 2), int(window / 2)), (0, 0)),
                       'edge')

    assert epis_padded.shape == (v_dim + window - 1,
                                 u_dim + window - 1), 'Padded epis array has wrong shape in function \'bilateral_median\'.'
    assert Me_padded.shape == (v_dim + window - 1,
                               u_dim + window - 1), 'Padded Mes array has wrong shape in function \'bilateral_median\'.'
    assert Ds_padded.shape == (v_dim + window - 1,
                               u_dim + window - 1), 'Padded Disps array has wrong shape in function \'bilateral_median\'.'


    smoothed_Disps = bilateral_median_inner(Ds_padded, epis_padded, Me_padded,
                                            M, np.array([v_dim, u_dim]),
                                            window, threshold)

    # Let's see if our results have reasonable meaning'
    assert smoothed_Disps.shape == (v_dim,
                                    u_dim), 'values_to_smooth_D output has wrong shape in function \'bilateral_median\'.'

    Ds[M] = smoothed_Disps[M]  # update values

    return Ds


@jit(nopython=True)
def bilateral_median_inner(Ds_padded, epis_padded, Me_padded, M, v_u_dim,
                           window, threshold):
    """
    Fast inner loop for the bilateral median filter.

    Parameters
    ----------
     Ds_padded : numpy.array [v+window,u+window].
        The set of already calculated disparities for each EPI at one scanline.
        The array was extended to avoid border problems.
    epis_padded : numpy.array [v+window,u+window]
        Set of all gray-value epis for one scanline. The array was extended
        to avoid border problems.
    Me_padded : numpy.array [v+window,u+window] of boolean.
        True means an edge was discovered for at that pixel.
    v_u_dim: numpy.array.
        The original v- and u-dimensions.
    M: numpy.array [u] boolean.
       Mask which values to consider.
    window : int, optional
        The window size for the bilateral median filter. As the window should
        be centered symetrically around the pixel to process the value shpuld
        be odd. For even numbers the next higher odd number is chosen.
    threshold : float, optional
        The threshold for the filter to determine if two EPI pixels are
        regarded as similar.

    Returns
    -------
    smoothed_Ds: numpy.array [v,u].
        The smoothed disparities at the current s-dimension.
    """

    v_dim = v_u_dim[0]
    u_dim = v_u_dim[1]

    smoothed_Ds = np.full((v_dim, u_dim), np.nan,dtype=np.float64)  # we collect all the values

    for v in range(v_dim):
        for u in range(u_dim):

            if not M[v, u]:
                continue

            v_hat = v + int(window // 2)
            u_hat = u + int(window // 2)

            values_to_smooth = []
            for v_ in range(v_hat - int(window // 2),
                            v_hat + int(window // 2) + 1):
                for u_ in range(u_hat - int(window // 2),
                                u_hat + int(window // 2) + 1):
                    assert not np.isnan(epis_padded[
                                            v_hat, u_hat]), 'NaN detected in EPI in function \'bilateral_median_inner\'.'
                    assert not np.isnan(epis_padded[
                                            v_, u_]), 'NaN detected in EPI in function \'bilateral_median_inner\'.'

                    diff = abs(epis_padded[v_hat, u_hat] - epis_padded[v_, u_])
                    if diff < threshold and Me_padded[v_, u_] and not np.isnan(
                            Ds_padded[
                                v_, u_]):  # There can be NaNs in other epis
                        values_to_smooth.append(Ds_padded[v_, u_])

            if len(values_to_smooth) > 0:
                tmp_array = np.zeros((len(values_to_smooth),), dtype=np.float64)
                for i in range(len(values_to_smooth)):
                    tmp_array[i] = values_to_smooth[i]
                smoothed_Ds[v, u] = np.median(tmp_array)

    return smoothed_Ds


def propagation(Ds, epi, R_bar_best, s_hat, threshold=0.1, DEBUG=False):
    """
    Propagate depth estimate for each entry in the current s-dimension.

    Parameters
    ----------
    Ds : numpy.array [v,s,u]
        The set of already calculated disparities for each EPI.
    epi : numpy.array [v,s,u]
        Set of all gray-value epis.
    R_bar_best : numpy.array [u]
        The best updated convergence radiance for each EPI pixel.
    s_hat : int
       The current scanline s to sample for.
    threshold : float, optional
        The threshold for the filter to determine if two EPI pixels are
        regarded as similar.
    DEBUG: boolean, optional
        Enable plotting to visualize the propagation process

    Returns
    -------
    Ds : numpy.array [v,s,u.
        The propagated disparities.
    plots : ndarray [v,s,u]
        If plotting was enabled epis with markings of the propagation process.
    """

    v_dim = epi.shape[0]
    s_dim = epi.shape[1]
    u_dim = epi.shape[2]
    # Check dimensions of input data
    assert Ds.shape == (v_dim, s_dim, u_dim,), 'Input Ds has wrong shape in function \'propagation\'.'
    assert epi.shape == (v_dim, s_dim, u_dim,), 'Input epi has wrong shape in function \'propagation\'.'
    assert R_bar_best.shape == (v_dim, u_dim,), 'Input R_bar_best has wrong shape in function \'propagation\'.'
    assert s_hat >= 0 and s_hat < s_dim, 's_hat not in range 0 - s-dim in function \'propagation\'.'

    if DEBUG:
        plots = epi.copy()
    else:
        plots = np.zeros(epi.shape)

    Disps, plot = propagation_inner(Ds, epi, R_bar_best, s_hat, threshold, plots, DEBUG=DEBUG)
    assert Ds.shape == (v_dim, s_dim, u_dim), 'Output Ds has wrong shape in function \'propagation\'.'

    return Ds, plot


@jit(nopython=True)
def propagation_inner(Ds, epi, R_bar_best, s_hat, threshold, plots, DEBUG=False):
    """
    Propagate depth estimate for each entry in the current s-dimension.

    Parameters
    ----------
    Ds : numpy.array [v,s,u]
        The set of already calculated disparities for each EPI.
    epi : numpy.array [v,s,u]
        Set of all gray-value epis.
    R_bar_best : numpy.array [u]
        The best updated convergence radiance for each EPI pixel.
    s_hat: int
       The current scanline s to sample for.
    threshold : float, optional
        The threshold for the filter to determine if two EPI pixels are
        regarded as similar.
    plots : numpy.array [d,s,u]
        If plotting was enabled epis to mark on the sampling process.
    DEBUG : boolean, optional
        Enable plotting to visualize the propagation process

    Returns
    -------
    Ds: numpy.array [v,s,u]
        The propagated disparities.
    plots: ndarray [v,s,u]
        If plotting was enabled epis with markings of the propagation process.
    """

    v_dim = epi.shape[0]
    s_dim = epi.shape[1]
    u_dim = epi.shape[2]

    # Go through all the entries in Disps
    for v in range(v_dim):
        for u in range(u_dim):
            if np.isnan(Ds[v, s_hat, u]):  # we need not to propagate NaNs
                continue

            # we need to check if a pixel can be reached by propagation
            # If there is an overlap both pixels are tried
            # go in u direction relative to current u
            for s in range(s_dim):
                x = u + (s_hat - s) * Ds[v, s_hat, u]
                x_floor = int(floor(x))
                x_ceil = int(ceil(x))

                # check the propagation conditions
                if x_floor >= 0 and x_floor < u_dim and np.isnan(
                        Ds[v, s, x_floor]):

                    diff = abs(epi[v, s, x_floor] - R_bar_best[v, u])
                    if diff < threshold:
                        Ds[v, s, x_floor] = Ds[v, s_hat, u]

                    if DEBUG:
                        plots[v, s, x_floor] = 1

                if x_ceil >= 0 and x_ceil < u_dim and np.isnan(
                        Ds[v, s, x_ceil]):

                    diff = abs(epi[v, s, x_ceil] - R_bar_best[v, u])
                    if diff < threshold:
                        Ds[v, s, x_ceil] = Ds[v, s_hat, u]

                    if DEBUG:
                        plots[v, s, x_ceil] = 1

    return Ds, plots



from setuptools import setup

setup(name='DisneyDisp',
      version='0.1.5',
      description='A python implementation of the \"Scene Reconstruction from High Spatio-Angular Resolution Light Fields\". For more information see https://www.disneyresearch.com/project/lightfields/.',
      url='https://github.com/manuSrep/DisneyDispPy.git',
      author='Marcel Gutsche, Manuel Tuschen',
      author_email='Manuel_Tuschen@web.de',
      license='GPL3 License',
      packages=['DisneyDisp'],
      install_requires=["scipy","numpy","scikit-image","matplotlib","numba","h5py","progressbar2","easyScripting"],
      zip_safe=False,
      entry_points={'console_scripts': ['disney=DisneyDisp.disney:main',
                                        'clif2lf=DisneyDisp.clif2lf:main',
                                        'dmap2hdf5=DisneyDisp.dmap2hdf5:main',
                                        'imgs2lf=DisneyDisp.imgs2lf:main',
                                        'lf2epi=DisneyDisp.lf2epi:main'],},
      )



#!/usr/bin/python
# -*- coding: utf8 -*-

"""
Disney Disparity.
:author: Manuel Tuschen
:date: 20.06.2016
:license: GPL3
"""

from __future__ import division, absolute_import, unicode_literals, print_function

import numpy as np
from skimage.transform import resize


def fine_to_course(D, course_res):
    """
    Downsample the disparities for the next resolution. Already calculated
    values are propagated.

    Parameters
    ----------
    D : numpy.array [v,s,u].
        Already calculated disparities for the finer resolution
    course_res : numpy.array  []
        The v,u dimension of the courser resolution.

    Returns
    -------
    D_ : numpy.array [v_,s,u_,]
        The update disparities for the courser resolution.
    """

    v_dim = D.shape[0]
    s_dim = D.shape[1]
    u_dim = D.shape[2]

    # Check dimensions of input data
    assert D.shape == (v_dim, s_dim,
                       u_dim), 'Input D has wrong input shape in function \'fine_to_course\'.'
    assert course_res[0] < v_dim, 'v-dimension of course_res larger than v_dim in function \'fine_to_course\'.'
    assert course_res[1] < u_dim, 'u-dimension of course_res larger than u_dim in function \'fine_to_course\'.'

    u_scale = course_res[1] / u_dim
    assert u_scale <= 1, 'The scaling factor of u-dimension is larger 1 in function \'fine_to_course\'.'

    D_ = resize(D, (course_res[0], s_dim, course_res[1]), order=0, mode='edge')
    D_ *= u_scale

    return D_


def course_to_fine(D_epi_fine, D_epi_course):
    """
    In a last step disparity estimates from lower resolution must be transferred
    to the higher resolutions if there was no value calculated yet.

    Parameters
    ----------
    D_epi_fine : numpy.array [v,s,u]
        Disparity estimates at the finer resolution.
    D_epi_course : numpy.array [v,s,u]
        Disparity estimates at the courser resolution.

    Returns
    -------
    D_epi_fine : numpy.array [v,s,u]
        Updated disparity estimates at the finer resolution.
    """

    v_dim_fine = D_epi_fine.shape[0]
    v_dim_course = D_epi_course.shape[0]
    u_dim_fine = D_epi_fine.shape[2]
    u_dim_course = D_epi_course.shape[2]
    s_dim = D_epi_course.shape[1]

    # Check dimensions of input data
    assert D_epi_fine.shape == (v_dim_fine, s_dim,
                                u_dim_fine), 'Input D_epi_fine has wrong input shape in function \'course_to_fine\'.'
    assert D_epi_course.shape == (v_dim_course, s_dim,
                                  u_dim_course), 'Input D_epi_course has wrong input shape in function \'course_to_fine\'.'
    assert v_dim_course < v_dim_fine, 'v-dimension of course image larger than v_dim of fine image in function \'course_to_fine\'.'
    assert u_dim_course < u_dim_fine, 'u-dimension of course image larger than v_dim of fine image in function \'course_to_fine\'.'

    u_scale = u_dim_course / u_dim_fine
    assert u_scale <= 1, 'The scaling factor of u-dimension is larger 1 in function \'course_to_fine\'.'

    D_tmp = np.zeros((v_dim_fine, s_dim, u_dim_fine), dtype=np.float64)

    for s in range(s_dim):
        D_tmp[:, s] = resize(D_epi_course[:, s], (v_dim_fine, u_dim_fine),
                             order=0, mode='edge')
        if not np.all(np.isnan(D_tmp[:, s])):
            assert not np.any(np.isnan(D_tmp[:,s])), 'NaN detected in D_tmp in function \'course_to_fine\'.'
    D_tmp /= u_scale

    replace = np.where(np.isnan(D_epi_fine))
    D_epi_fine[replace] = D_tmp[replace]
    if not np.all(np.isnan(D_epi_fine[:, s])):
        assert not np.any(np.isnan(D_epi_fine[:,s])), 'NaN detected in output D_epi_fine in function \'course_to_fine\'.'

    return D_epi_fine
